from .audio_identifier import SpeakerIdentifier, db_speaker_id
from .logger_config import get_log_folder_path

dir_root = get_log_folder_path()
speaker_pipeline = SpeakerIdentifier(persist_directory=dir_root)


def get_speaker_system_appendix():
    out = "You can hear anyone in room speaking. That is why you have been given an annotation of who is speaking. " \
          "In the form of:\n'[Unknown Speaker X]:  ...' or '[John Doe]:  ...' or '[Jane]:  ...'\n Do not include " \
          "any speaker annotation or any other brackets in your response. It is your job to keep track of who is " \
          "speaking and to respond to the correct person. If you are unsure who is speaking, ask for clarification. " \
          "The only way the system can know who is speaking is if you tell it to associate an Unknown Speaker with a " \
          "name by asking the user(s) to provide a name and storing it with the correct function. The user(s) do not " \
          "know what 'Unknown Speaker X' number they are. Do not ask them to provide a name for a specific " \
          "'Unknown Speaker X' number. Instead, ask them to provide a name for the person that is speaking, " \
          "who spoke last, or who said something specific. If you are unsure who is speaking, ask for clarification. " \
          "This means you CAN recognize voices, just trust your annotations (remember the same person may talk with " \
          "multiple voices)."
    return out


def store_name_for_unknown_speaker(unknown_speaker_id="", name=""):
    if name == "":
        raise ValueError("The name cannot be empty.")
    if name.lower().find("unknown") != -1:
        raise ValueError("The name cannot contain 'Unknown'.")
    if name.lower().find("speaker") != -1:
        raise ValueError("The name cannot contain 'Speaker'.")
    if unknown_speaker_id == "":
        raise ValueError("The unknown_speaker_id cannot be empty.")
    if unknown_speaker_id.find("Unknown Speaker ") == -1:
        if unknown_speaker_id.isnumeric():
            unknown_speaker_id = "Unknown Speaker "+unknown_speaker_id
        else:
            raise ValueError("The unknown_speaker_id must start with 'Unknown Speaker '.")
    try:
        embedding = speaker_pipeline.get_unknown_embedding(unknown_speaker_id)
    except TypeError:
        embedding = None
    if embedding is None:
        raise ValueError("The unknown_speaker_id does not exist.")
    speaker_pipeline.add_known_speaker(embedding, name)
    speaker_pipeline.remove_unknown_speaker(unknown_speaker_id)
    return True


def store_name_for_unknown_speaker_documentation():
    schema = {"type": "function",
              "function": {
                    "name": "store_name_for_unknown_speaker",
                    "description": "Stores a name for an unknown speaker.",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "unknown_speaker_id": {
                                "type": "string",
                                "description": "The unknown speaker id to store the name for ex. 'Unknown Speaker X' "
                                               "or just 'X'.",
                            },
                            "name": {
                                "type": "string",
                                "description": "The name to store for the unknown speaker.",
                            },
                        },
                        "required": ["unknown_speaker_id", "name"],
                    },
                }
            }
    examples = 'Examples:\n {"function_name": "store_name_for_unknown_speaker", "parameters": {"unknown_speaker_id": ' \
               '"Unknown Speaker 1", "name": "John Doe"}}\n{"function_name": "store_name_for_unknown_speaker", ' \
               '"parameters": {"unknown_speaker_id": "Unknown Speaker 2", "name": "Jane"}}\n'
    return schema, examples


def remove_name_of_known_speaker(name=""):
    if name == "":
        raise ValueError("The name cannot be empty.")
    if name.lower().find("unknown") != -1:
        raise ValueError("The name cannot contain 'Unknown'.")
    speaker_pipeline.remove_known_speaker(name)
    return True


def remove_name_of_known_speaker_documentation():
    schema = {"type": "function",
              "function": {
                    "name": "remove_name_of_known_speaker",
                    "description": "Removes a name of a known speaker.",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "name": {
                                "type": "string",
                                "description": "The name of the known speaker to remove.",
                            },
                        },
                        "required": ["name"],
                    },
                }
            }
    examples = 'Examples:\n {"function_name": "remove_name_of_known_speaker", "parameters": {"name": "John Doe"}}\n' \
               '{"function_name": "remove_name_of_known_speaker", "parameters": {"name": "Jane"}}\n'
    return schema, examples


def get_speaker_function_info():
    return {"store_name_for_unknown_speaker": {"function": store_name_for_unknown_speaker,
                                               "schema": store_name_for_unknown_speaker_documentation()[0],
                                               "examples": store_name_for_unknown_speaker_documentation()[1]},
            "remove_name_of_known_speaker": {"function": remove_name_of_known_speaker,
                                             "schema": remove_name_of_known_speaker_documentation()[0],
                                             "examples": remove_name_of_known_speaker_documentation()[1]},
            }


info = get_speaker_function_info()
function_list = []
function_examples = []
for val in info.values():
    function_list.append(val["schema"])
    function_examples.append(val["examples"])


def get_speaker_function_list():
    return function_list


def get_speaker_function_examples():
    return function_examples



import os
import time
import fcntl
import atexit

import faiss
import numpy as np
import sqlite3


def normalize_embedding(embedding):
    """Normalize an embedding vector to unit length."""
    return embedding / np.linalg.norm(embedding)


class LocalAudioDB:

    def __init__(self, path, embedding_dim=256):
        if not os.path.exists(path):
            os.mkdir(path)
        db_path = os.path.join(path, 'embeddings_index.db')
        lock_path = os.path.join(path, 'localaudiodb.lock')
        self.db_path = db_path
        self.lock_path = lock_path
        self.embedding_dim = embedding_dim
        self.conn = sqlite3.connect(db_path, check_same_thread=False)
        self.cursor = self.conn.cursor()
        self.initialize_db()
        self.index = faiss.IndexFlatL2(embedding_dim)
        self.last_update_time = self.get_last_modified_time()
        if self.last_update_time is None:
            self.acquire_lock(check_for_updates=False)
            try:
                self.update_last_modified_time()
            finally:
                self.release_lock()
        self.lock_file = None
        self.reload_index_from_db()
        atexit.register(self.close)
        atexit.register(self.try_release_lock)

    def initialize_db(self):
        self.cursor.execute('''CREATE TABLE IF NOT EXISTS embedding_index 
                               (id TEXT PRIMARY KEY, embedding BLOB, normal_embedding BLOB)''')
        self.conn.commit()
        self.cursor.execute('''CREATE TABLE IF NOT EXISTS used_ids
                                 (id TEXT PRIMARY KEY)''')
        self.conn.commit()
        self.cursor.execute('''CREATE TABLE IF NOT EXISTS last_update_time
                                    (timestamp REAL)''')
        self.conn.commit()

    def acquire_lock(self, check_for_updates=True):
        self.lock_file = open(self.lock_path, 'w')
        fcntl.flock(self.lock_file, fcntl.LOCK_EX)
        if check_for_updates:
            self.check_for_updates()

    def release_lock(self):
        fcntl.flock(self.lock_file, fcntl.LOCK_UN)
        self.lock_file.close()

    def try_release_lock(self):
        try:
            self.release_lock()
        except:
            pass

    def update_last_modified_time(self):
        # Must have lock
        self.cursor.execute("DELETE FROM last_update_time")
        self.conn.commit()
        self.cursor.execute("INSERT INTO last_update_time (timestamp) VALUES (?)", (time.time(),))
        self.conn.commit()
        self.last_update_time = time.time()

    def get_last_modified_time(self):
        self.cursor.execute("SELECT timestamp FROM last_update_time")
        result = self.cursor.fetchone()
        return result[0] if result else None

    def check_for_updates(self):
        current_update_time = self.get_last_modified_time()
        if current_update_time > self.last_update_time:
            self.reload_index_from_db()
            self.last_update_time = current_update_time
        else:
            self.cursor.execute("SELECT COUNT(*) FROM embedding_index")
            result = self.cursor.fetchone()
            max_index = result[0] if result else None
            if max_index is not None and max_index != self.index.ntotal:
                self.try_release_lock()
                raise Exception("FAISS and SQLite database are out of sync: " +
                                str(max_index) + " " + str(self.index.ntotal))

    def get_largest_id(self):
        # check max id between used_ids and embedding_index
        self.cursor.execute("SELECT MAX(id) AS max_id FROM ("
                            " SELECT id FROM used_ids UNION ALL SELECT id FROM embedding_index);")
        result = self.cursor.fetchone()
        return result[0] if result else None

    def add_embedding(self, audio_id, embedding):
        self.acquire_lock()
        try:
            normal_embedding = normalize_embedding(embedding)
            embedding_blob = embedding.tobytes()
            normal_embedding_blob = normal_embedding.tobytes()
            self.cursor.execute("INSERT INTO embedding_index (id, embedding, normal_embedding) "
                                "VALUES (?, ?, ?)",
                                (str(audio_id), embedding_blob, normal_embedding_blob))
            self.conn.commit()
            self.update_last_modified_time()
            self.reload_index_from_db()
        finally:
            self.release_lock()

    def get_embedding(self, audio_id):
        self.check_for_updates()
        self.cursor.execute("SELECT id, embedding FROM embedding_index WHERE id = ?", (str(audio_id),))
        id, embedding_blob = self.cursor.fetchone()
        embedding = np.frombuffer(embedding_blob, dtype=np.float32)
        return id, embedding

    def query_embeddings(self, audio_id_snippet):
        self.check_for_updates()
        self.cursor.execute("SELECT id, embedding FROM embedding_index WHERE id "
                            "LIKE ?", ('%' + audio_id_snippet + '%',))
        results = self.cursor.fetchall()
        if results:
            for i, result in enumerate(results):
                id, embedding_blob = result
                embedding = np.frombuffer(embedding_blob, dtype=np.float32)
                results[i] = (id, embedding)
            return results

    def remove_embedding(self, audio_id):
        self.acquire_lock()
        try:
            self.cursor.execute("DELETE FROM embedding_index WHERE id = ?",
                                (str(audio_id),))
            self.conn.commit()
            self.cursor.execute("INSERT INTO used_ids (id) VALUES (?)", (str(audio_id),))
            self.conn.commit()

            self.update_last_modified_time()
            self.reload_index_from_db()
        finally:
            self.release_lock()
        return True

    def search_embeddings(self, query_embedding, k=5):
        self.check_for_updates()
        query_embedding = normalize_embedding(query_embedding)
        distances, indices = self.index.search(np.array([query_embedding]), k)
        return distances, indices

    def find_closest_embedding(self, query_embedding, retry=3):
        self.acquire_lock()
        try:
            if self.index.ntotal == 0:
                return None, None, None
            normal_query_embedding = normalize_embedding(query_embedding)

            distances, indices = self.index.search(np.array([normal_query_embedding]), 1)

            closest_index = indices[0][0]
            closest_distance = distances[0][0]

            # Retrieve the corresponding ID by sorting the IDs in ascending order and selecting the ID at the index
            self.cursor.execute("SELECT id, normal_embedding FROM embedding_index ORDER BY id ASC LIMIT 1 OFFSET ?-1;",
                                (closest_index+1,))
            result = self.cursor.fetchone()
            if result:
                closest_id, embedding_blob = result
                closest_embedding = np.frombuffer(embedding_blob, dtype=np.float32)
                return closest_id, closest_embedding, closest_distance
            else:
                if self.index.ntotal == 1:
                    return None, None, None
                if retry > 0:
                    self.release_lock()
                    self.find_closest_embedding(query_embedding, retry=retry-1)
                else:
                    self.try_release_lock()
                    raise Exception("FAISS index and SQLite database are out of sync.")
        finally:
            self.release_lock()

    def reload_index_from_db(self):
        self.cursor.execute("SELECT normal_embedding FROM embedding_index ORDER BY id ASC")
        embeddings = self.cursor.fetchall()
        if embeddings:
            embeddings_array = np.array([np.frombuffer(e[0], dtype=np.float32) for e in embeddings])
            self.index = faiss.IndexFlatL2(self.embedding_dim)
            self.index.add(embeddings_array)

    def close(self):
        self.conn.close()





import wave
import time
from io import BytesIO
from numpy import frombuffer, int16, average, isnan, concatenate

from .logger_config import get_logger

logger = get_logger()


def prep_mic(duration: float = 2.0) -> int:
    """
    Prepare the microphone for listening by adjusting it for ambient noise.
    :param duration: The duration to adjust the microphone for.
    :return: Adjusted energy
    """
    import sounddevice as sd
    frames_per_buffer = 1024
    sample_rate = 16000
    seconds_per_buffer = frames_per_buffer / sample_rate
    elapsed_time = 0
    total_volume = 0
    count = 0

    with sd.InputStream(samplerate=sample_rate, channels=1, dtype='int16') as stream:
        stream.start()
        while elapsed_time < duration:
            chunk, overflowed = stream.read(frames_per_buffer)
            if overflowed:
                logger.warning("Input overflow detected")
            if not isnan(chunk).any():  # Check for NaN values
                chunk = chunk.astype(int16).tobytes()
                volume = average(abs(frombuffer(chunk, dtype=int16)))
                if not isnan(volume):
                    total_volume += volume
                    count += 1
            elapsed_time += seconds_per_buffer

    avgen = total_volume/count
    logger.info(f"Final Adjusted Microphone Energy: {avgen}")
    return int(avgen)


def listen_to_user(energy, silence_threshold=.5, maximum_seconds=120) -> BytesIO:
    """
    Listen to the user and record their speech, stopping when there's silence.

    :return: The recorded audio data.
    :rtype: BytesIO
    """
    import sounddevice as sd
    # Initialize PyAudio and create a stream
    sample_rate = 16000
    frames_per_buffer = 4096
    seconds_per_buffer = frames_per_buffer / sample_rate
    energy_factor = 1.2
    energy = energy*energy_factor
    with sd.InputStream(samplerate=sample_rate, channels=1, dtype='int16') as stream:

        audio_data = None
        window_size = round(silence_threshold * 2 / seconds_per_buffer)
        volume_buffer = []
        silence_duration = 0
        start = time.time()

        while time.time() - start < maximum_seconds:
            # Read a chunk of audio data from the stream

            chunk, overflowed = stream.read(frames_per_buffer)
            if overflowed:
                logger.warning("Input overflow detected")
            chunk = chunk.astype(int16)
            if audio_data is None:
                audio_data = chunk.flatten()
            else:
                audio_data = concatenate((audio_data, chunk.flatten()))

            # Calculate the volume of the audio data in the chunk
            volume_buffer.append(average(abs(chunk)))

            # Keep track of the average volume over the last window_size chunks
            if len(volume_buffer) > window_size:
                volume_buffer.pop(0)
            avg_volume = average(volume_buffer)
            # Keep track of the duration of silence
            if avg_volume < energy:
                silence_duration += seconds_per_buffer
            else:
                silence_duration = 0

            # Stop recording if there is enough silence
            if silence_duration > silence_threshold and time.time() - start > 2:
                break

        # Close the stream and terminate PyAudio

        if time.time() - start < 3:
            return None

        # Create an AudioData object from the recorded audio data
        audio_bytes_wave = BytesIO()
        with wave.open(audio_bytes_wave, 'wb') as wf:
            wf.setnchannels(1)
            wf.setsampwidth(2)  # 2 bytes per sample since dtype='int16'
            wf.setframerate(sample_rate)
            wf.writeframes(audio_data)

        return audio_bytes_wave


def audio_capture_process(audio_queue, speaking, stop_event):
    """
    Captures audio from the microphone and puts it in the audio queue.
    :param audio_queue:
    :param speaking:
    :param stop_event:
    :return: None
    """
    try:
        first = True
        level = prep_mic()
        while stop_event.is_set() is False:
            if not speaking.is_set():
                audio_data = listen_to_user(level)
                if not speaking.is_set() and not stop_event.is_set() and audio_data is not None:
                    audio_queue.put((audio_data, time.time()))
                if first:
                    first = False
                    audio_queue.put((None, time.time()))
    except KeyboardInterrupt:
        return


import os
from .logger_config import get_log_folder_path, get_logger
logger = get_logger()


dir_root = get_log_folder_path()

dir_path = os.path.join(dir_root, "whisper_models")
if not os.path.exists(dir_path):
    os.mkdir(dir_path)


def recognition_process(audio_queue, text_queue, speaking, stop_event):
    try:
        from .audio_identifier import SpeakerIdentifier
        speaker_pipeline = SpeakerIdentifier(persist_directory=dir_root)
        speaker_pipeline.speedup_if_able()
        while stop_event.is_set() is False:
            if audio_queue.empty():
                speaking.wait(timeout=0.2)
            else:
                audio_data = audio_queue.get()
                if not speaking.is_set() and not stop_event.is_set() and audio_data is not None:
                    speakers = speaker_pipeline.get_speakers(audio_data)
                    text_queue.put(speakers)
    except KeyboardInterrupt:
        pass



import copy
import datetime
import json
import os
import re
import time
import warnings
import uuid
from typing import List
import chromadb
from chromadb.config import Settings


def get_time() -> tuple:
    """
    Get the current time as a string.

    :return: A tuple containing two strings: the current time and the current UTC time.
    :rtype: tuple
    """
    return (
        f"On {datetime.datetime.now().strftime('%A, %B %-d, %Y at %-I:%M %p')}: ",
        str(datetime.datetime.utcnow()),
    )


def convert_utc_to_local(utc_time: str) -> str:
    """
    Convert a UTC time string to a local time string.

    :param utc_time: A string containing a UTC time.
    :type utc_time: str
    :return: A string containing the local time.
    :rtype: str
    """
    utc_time = datetime.datetime.strptime(utc_time, "%Y-%m-%d %H:%M:%S.%f")
    local_time = utc_time.replace(tzinfo=datetime.timezone.utc).astimezone(tz=None)
    return local_time.strftime("%A, %B %-d, %Y at %-I:%M %p")


def _strip_entry(entry: dict or list):
    """
    Remove all fields from the entry dictionary except 'role' and 'content'.

    :param entry: A dictionary containing a conversation entry.
    :type entry: dict
    :return: A new dictionary containing only the 'role' and 'content' fields from the original entry.
    :rtype: dict
    """
    if isinstance(entry, list):
        new = []
        for el in entry:
            new.append(_strip_entry(el))
        return new
    else:
        if "name" in entry:
            return {"role": entry["role"], "content": entry["content"], "name": entry["name"]}
        return {"role": entry["role"], "content": entry["content"]}


class AssistantHistory:
    """
    A class to manage the Assistant's conversation history, including storing, reducing,
    and gathering conversation context for future queries.

    :param username: The username of the current user.
    :type username: str
    :param system: The system message.
    :type system: str
    :param tokenizer: A function to tokenize a string.
    :type tokenizer: function
    :param summarizer: A function to summarize a string.
    :type summarizer: function
    :param max_tokens: The maximum number of tokens allowed for the conversation history.
    :type max_tokens: int
    :param embedder: A function to embed a string.
    :type embedder: function, optional
    :param persist_directory: The directory to store the database in.
    :type persist_directory: str, optional
    :param model_injection: Whether to inject the model name into the history.
    :type model_injection: bool, optional
    :param time_injection: Whether to inject the time into the history.
    :type time_injection: bool, optional
    """

    def __init__(
        self,
        username: str,
        system: str,
        tokenizer: callable,
        summarizer: callable,
        max_tokens: int,
        summary_max_tokens: int,
        embedder: callable = None,
        persist_directory: str = "database",
        model_injection: bool = True,
        time_injection: bool = True,
    ):
        """
        Initialize an instance of AssistantHistory.
        """
        self.model_injection = model_injection
        self.time_injection = time_injection
        self.persist_directory = persist_directory
        self.client = chromadb.PersistentClient(path=self.persist_directory,
                                                settings=Settings(anonymized_telemetry=False))
        self.remove_broken_id_updates()
        self.to_summarize = []
        # Load long-term memory from disk, or create a new LTM file if it doesn't exist.
        if os.path.exists(os.path.join(self.persist_directory, "AssistantHistoryLTM.json")):
            with open(os.path.join(self.persist_directory, "AssistantHistoryLTM.json"), "r") as f:
                ltm = json.load(f)
        else:
            ltm = {"long_term_memory": ""}
        self.long_term_memory = ltm["long_term_memory"]

        # Set instance variables based on input parameters.
        self.username = username
        self.fixed_user = username + "'" if username[-1] == "s" else username + "'s"
        self.system_raw = system
        self.embedder = embedder
        self.max_tokens = max_tokens
        self.summary_max_tokens = summary_max_tokens
        self.tokenizer = tokenizer
        self.summarizer = summarizer
        self.last_context = None
        # Create history and summaries collections using the chromadb client.
        if self.embedder:
            self.history = self.client.get_or_create_collection(
                name="history", embedding_function=self.embedder
            )
            self.summaries = self.client.get_or_create_collection(
                name="summaries", embedding_function=self.embedder
            )
        else:
            self.history = self.client.get_or_create_collection(name="history")
            self.summaries = self.client.get_or_create_collection(name="summaries")

        if int(self.get_current_id()) == 0 and len(self.history.peek()['ids']) > 0:
            warnings.warn("Chat assistant: history database needs metadate update. Updating...")
            updating = True
            seed = "updates"
            i = 1
            while updating:
                if not self.get_history_from_id_and_earlier(int(self.get_current_id()) + i, n_results=1):
                    updating = False
                else:
                    ignore = self.create_id(seed)
                    i += 1
            self.resolve_id(seed)

        if int(self.get_current_id(summary=True)) == 0 and len(self.summaries.peek()['ids']) > 0:
            warnings.warn("Chat assistant: summary database needs metadate update. Updating...")
            updating = True
            seed = "updates"
            i = 1
            while updating:
                if not self.get_summary_from_id_and_earlier(int(self.get_current_id(summary=True))+i, n_results=1):
                    updating = False
                else:
                    ignore = self.create_id(seed, summary=True)
                    i += 1
            self.resolve_id(seed, summary=True)

        if int(self.get_current_id()) > 1:
            # Check whether the next summary ID is in the history database.
            not_summarized = []
            last_batch = self.get_history_from_last_batch()
            last_batch_id = last_batch[0]["batch_id"]
            last_summary = self.get_summary_from_id_and_earlier(n_results=1)
            refresh = False
            if len(last_summary) > 0:
                last_summary_batch_id = last_summary[0]["batch_id"]
                if last_summary_batch_id < last_batch_id:
                    not_summarized = self.get_batches(list(range(int(last_summary_batch_id) + 1, int(last_batch_id))))
                    refresh = True
            else:
                not_summarized = last_batch
                refresh = True
            if refresh:
                self.to_summarize += not_summarized
                self.reduce_context()

    def remove_broken_id_updates(self):
        """
        Remove any broken ID updates.
        """
        # list files in persist_directory
        files = os.listdir(self.persist_directory)
        for fn in files:
            if fn.endswith(".tmp"):
                os.remove(os.path.join(self.persist_directory, fn))
                warnings.warn("Chat assistant: was interrupted during an update. Removing broken update file.")

    def get_current_id(self, summary=False):
        """
        Get the current set ID.

        :return: The current set ID.
        :rtype: str
        """
        mode = "summary" if summary else "chat"
        path = os.path.join(self.persist_directory, ".current_id_"+mode+ ".maxdb")
        if os.path.exists(path):
            with open(path, "r") as f:
                return str(f.read())
        else:
            with open(path, "w") as f:
                f.write("0")
            return '0'

    def create_id(self, seed, summary=False):
        """
        Get the current set ID.

        :return: The current set ID.
        :rtype: str
        """
        seed = os.path.basename(seed)
        assert len(seed) > 4
        mode = "summary" if summary else "chat"
        path = os.path.join(self.persist_directory, ".in_process_id_"+mode + "_" + seed + ".maxdb")
        if os.path.exists(path):
            with open(path, "r") as f:
                current = f.read()
        else:
            test = os.listdir(os.path.join(self.persist_directory))
            test = [x for x in test if os.path.basename(x).startswith(".in_process_id_"+mode)]
            if len(test) > 0:
                for fn in test:
                    os.remove(os.path.join(self.persist_directory, fn))
                raise Exception("Chat assistant: multiple in process id files found.")
            current = self.get_current_id(summary=summary)
        next_id = str(int(current)+1)
        with open(path+".tmp", "w") as f:
            f.write(next_id)
        os.replace(path+".tmp", path)
        return next_id

    def resolve_id(self, seed, summary=False):
        mode = "summary" if summary else "chat"
        inp_path = os.path.join(self.persist_directory, ".in_process_id_" + mode + "_" + seed + ".maxdb")
        out_path = os.path.join(self.persist_directory, ".current_id_" + mode + ".maxdb")
        os.replace(inp_path, out_path)

    def save_ltm(self):
        """
        Save the long term memory to a file.
        """
        with open(os.path.join(self.persist_directory, "AssistantHistoryLTM.json"), "w") as f:
            json.dump({"long_term_memory": self.long_term_memory}, f)

    def update_ltm(self):
        self.long_term_memory = self.summarizer(
            self.gather_context("", only_summaries=True,
                                max_tokens=(self.summary_max_tokens -
                                            self.count_tokens_text(self.long_term_memory))))["content"]
        self.save_ltm()

    def count_tokens_text(self, text: str) -> int:
        """
        Count the number of tokens in the given text.

        :param text: A string containing the text to count tokens for.
        :type text: str
        :return: The number of tokens in the given text.
        :rtype: int
        """
        return len(self.tokenizer(text))

    def count_tokens_context(self, ls: list) -> int:
        """
        Count the total number of tokens in a list of conversation entries.

        :param ls: A list of conversation entries.
        :type ls: list
        :return: The total number of tokens in the given list of entries.
        :rtype: int
        """
        total = 0
        for el in ls:
            if isinstance(el, list):
                for e in el:
                    for k in e.keys():
                        test = e[k]
                        if isinstance(test, str):
                            total += self.count_tokens_text(test)
            else:
                for k in el.keys():
                    test = el[k]
                    if isinstance(test, str):
                        total += self.count_tokens_text(test)
        return total

    def add_context(self, context: list) -> None:
        """
        Add query and response to the conversation history.

        :param context: A list containing the query and response.
        :type context: list
        """
        time_str, utc_time = get_time()
        documents = []
        first_id = None
        ids = []
        seed = str(uuid.uuid4())
        for i in range(len(context)):
            context[i]['id'] = self.create_id(seed)
            ids.append(context[i]['id'])
            if first_id is None:
                first_id = context[i]['id']
            context[i]['utc_time'] = utc_time
            context[i]['batch_id'] = first_id
            if context[i]['role'] == 'assistant' and context[i].get('model', None) is not None:
                context[i]['model'] = context[i]['model']
            context[i]['num_tokens'] = self.count_tokens_text(context[i]['content'])
            if self.model_injection and context[i].get('model', None) is not None:
                context[i]['content'] = "Source AI Model: " + context[i]['model'] + " - " + context[i]['content']
            documents.append(context[i]['content'])

        metadata = copy.deepcopy(context)
        if self.embedder:
            ebeddings = []
            for i in range(len(metadata)):
                ebeddings.append(self.embedder(metadata[i]['content']))
                del metadata[i]['content']

            self.history.add(
                embeddings=ebeddings,
                metadatas=metadata,
                documents=documents,
                ids=ids,
            )
        else:
            self.history.add(
                metadatas=metadata,
                documents=documents,
                ids=ids,
            )
        self.resolve_id(seed)
        self.to_summarize + context

    def get_system(self) -> dict:
        """
        Generate a system message containing user's AI Assistant's name and the current date time.

        :return: A dictionary containing the role and content of the system message.
        :rtype: dict
        """
        system_raw = self.system_raw
        system = re.sub("FIXED_USER_INJECTION", self.fixed_user, system_raw)
        system = re.sub("DATETIME_INJECTION", datetime.datetime.now().strftime("%A, %B %d, %Y at %I:%M %p"), system)
        system = re.sub("LONG_TERM_MEMORY_INJECTION", self.long_term_memory, system)
        return {"role": "system", "content": system}

    def reduce_context(self) -> None:
        """
        Reduce the conversation history by summarizing it.
        """
        if self.to_summarize is None:
            return
        if len(self.to_summarize) == 0:
            return
        batch_ids = []
        for elem in self.to_summarize:
            batch_ids.append(elem['batch_id'])
        for batch in batch_ids:
            source_ids = []
            batch_entries = []
            for el in self.to_summarize:
                if el['batch_id'] == batch:
                    source_ids.append(el['id'])
                    batch_entries.append(el)
            to_reduce = _strip_entry(batch_entries)
            new_summary = self.summarizer(to_reduce)
            new_summary_metadata = {"role": "assistant",
                                    "source_ids": ",".join(source_ids),
                                    "batch_id": batch,
                                    "utc_time": batch_entries[0]['utc_time'],
                                    "num_tokens": self.count_tokens_text(new_summary['content'])}
            seed = str(uuid.uuid4())
            if self.embedder:
                self.summaries.add(
                    embeddings=[self.embedder(new_summary['content'])],
                    metadatas=[new_summary_metadata.copy()],
                    documents=[new_summary['content']],
                    ids=[self.create_id(seed, summary=True)],
                )
            else:
                self.summaries.add(
                    metadatas=[new_summary_metadata.copy()],
                    documents=[new_summary['content']],
                    ids=[self.create_id(seed, summary=True)],
                )
            self.resolve_id(seed, summary=True)
            time.sleep(0.1)
        self.update_ltm()
        self.to_summarize = []

    def gather_context(self, query: str or list, minimum_recent_history_length: int = 2, max_tokens: int = None,
                       only_summaries: bool = False, only_necessary_fields: bool = True,
                       distance_cut_off: float = None, query_n_max: int = 3, verbose: bool = False) -> List[dict]:
        """
        Gathers relevant context for a given query from the chat assistant's history.

        :param query: The input query for which context is to be gathered
        :type query: str or list
        :param minimum_recent_history_length: The minimum number of recent history entries to include, defaults to 2
        :type minimum_recent_history_length: int, optional
        :param max_tokens: The maximum number of tokens allowed in the combined context, defaults to 2500
        :type max_tokens: int, optional
        :param only_summaries: Whether to only include summaries in the context, defaults to False
        :type only_summaries: bool, optional
        :param only_necessary_fields: Whether to only include 'role' and 'content' in context entries, defaults to True
        :type only_necessary_fields: bool, optional
        :param distance_cut_off: The maximum distance between the query and the context entry, defaults to None
        :type distance_cut_off: float, optional
        :param query_n_max: The maximum number results to return from full context query, defaults to 30
        :type query_n_max: int, optional
        :param verbose: Whether to print out context info
        :return: A list of relevant context entries for the given query
        :rtype: List[dict]
        """
        info = {"recent_context": {"ids": [], "num_tokens": 0},
                "history_query": {"ids": [], "num_tokens": 0},
                "summaries_queries": {"ids": [], "num_tokens": 0},
                "summaries": {"ids": [], "num_tokens": 0}}
        system_message = [self.get_system()]
        if isinstance(query, list):
            query_tokens = self.count_tokens_context(query+system_message)
            query_str = ""
            for el in query:
                query_str = query_str + "\n\n" + el["content"]
            query = query_str
        else:
            query_tokens = self.count_tokens_text(query)+self.count_tokens_context(system_message)

        if max_tokens is None:
            max_tokens = int(0.85 * self.max_tokens)
        if not only_summaries:
            context_list = []
            id_added = []
            summary_ids = []
            token_count = query_tokens

            # Add the most recent history entries
            recent_history_length = 0
            while recent_history_length < minimum_recent_history_length:
                if recent_history_length == 0:
                    batch_context = self.get_history_from_last_batch()
                    if batch_context is None:
                        break
                else:
                    batch_context = self.get_batch_before(batch_id)
                    if batch_context is None:
                        break
                batch_id = batch_context[0]["batch_id"]
                batch_context.reverse()
                for entry in batch_context:
                    if token_count + entry["num_tokens"] > max_tokens:
                        break
                    context_list.insert(0, entry)
                    tid = entry["id"]
                    id_added.append(tid)
                    token_count += entry["num_tokens"]
                    info["recent_context"]["ids"].append(tid)
                    info["recent_context"]["num_tokens"] += entry["num_tokens"]
                recent_history_length += 1

            # If the context is too short, query the full history
            current_id = int(self.get_current_id())
            if token_count < max_tokens and current_id > 1:
                query_size = query_n_max
                if query_size > current_id:
                    query_size = current_id
                if query_size > 0:
                    if self.embedder:
                        query_results = self.history.query(query_embeddings=[self.embedder(query)], n_results=query_size)
                    else:
                        query_results = self.history.query(query_texts=[query], n_results=query_size)
                    for tid in query_results["ids"][0][:query_size]:
                        if tid not in id_added:
                            result_pos = query_results["ids"][0].index(tid)
                            entry = query_results["metadatas"][0][result_pos]
                            if token_count + entry["num_tokens"] > max_tokens:
                                break
                            if distance_cut_off is not None:
                                if query_results["distances"][0][result_pos] < distance_cut_off:
                                    break
                            batch_data = self.get_batches(entry["batch_id"])
                            batch_data.reverse()
                            for item in batch_data:
                                if token_count + item["num_tokens"] > max_tokens:
                                    break
                                if item['id'] not in id_added:
                                    context_list.insert(0, item)
                                    token_count += item["num_tokens"]
                                    id_added.append(item['id'])
                                    info["history_query"]["ids"].append(item['id'])
                                    info["history_query"]["num_tokens"] += item["num_tokens"]

                    # If the context is still too short, query the summaries
                    if token_count < max_tokens:
                        query_size = query_n_max
                        current_summary_id = int(self.get_current_id(summary=True))
                        if query_size > current_summary_id:
                            query_size = current_summary_id
                        if query_size > 0:
                            if self.embedder:
                                query_summaries = self.summaries.query(query_embeddings=[self.embedder(query)],
                                                                       n_results=query_size)
                            else:
                                query_summaries = self.summaries.query(query_texts=[query], n_results=query_size)
                            for tid in query_summaries["ids"][0]:
                                if tid not in summary_ids:
                                    result_pos = query_summaries["ids"][0].index(tid)
                                    entry = query_summaries["metadatas"][0][result_pos]
                                    if token_count + entry["num_tokens"] > max_tokens:
                                        break
                                    if distance_cut_off is not None:
                                        if query_summaries["distances"][0][result_pos] < distance_cut_off:
                                            break
                                    if not (entry["source_ids"].split(",")[0] in id_added and
                                            entry["source_ids"].split(",")[1] in id_added):
                                        entry["content"] = query_summaries["documents"][0][result_pos]
                                        summary_ids.append(tid)
                                        context_list.insert(0, entry)
                                        token_count += entry["num_tokens"]
                                        info["summaries_queries"]["ids"].append(tid)


            else:
                context_list = []
                summary_ids = []
                id_added = []
                token_count = query_tokens
        else:
            context_list = []
            summary_ids = []
            id_added = []
            token_count = query_tokens

        # Add the summaries if there is any space left
        current_summary_id = int(self.get_current_id(summary=True))
        while current_summary_id > 0 and token_count < max_tokens:
            if current_summary_id not in summary_ids:
                result = self.summaries.get(ids=str(current_summary_id), include=['documents', 'metadatas'])
                entry = result["metadatas"][0]
                entry["content"] = result["documents"][0]
                if token_count + entry["num_tokens"] > max_tokens:
                    break
                if not (entry["source_ids"].split(",")[0] in id_added and
                        entry["source_ids"].split(",")[1] in id_added):
                    context_list.insert(0, entry)
                    summary_ids.append(result["ids"][0])
                    token_count += entry["num_tokens"]
                    info["summaries"]["ids"].append(result["ids"][0])
                    info["summaries"]["num_tokens"] += entry["num_tokens"]
            current_summary_id -= 1

        assert token_count <= max_tokens

        if only_necessary_fields:
            context_list = [_strip_entry(entry) for entry in context_list]

        self.last_context = system_message + context_list

        if verbose:
            from pprint import pprint
            pprint(info)

        return self.last_context

    def get_history(self):
        """
        Returns the history of the chat assistant

        :return: The complete history
        :rtype: list
        """
        return self.history

    def get_history_from_id_and_earlier(self, hid=None, n_results=10):
        """
        Returns the history of the chat assistant from a given id and earlier

        :param hid: The id to start from
        :type hid: int
        :param n_results: The number of results to return
        :type n_results: int
        :return: The history
        :rtype: list
        """
        if hid is None:
            hid = int(self.get_current_id())
        else:
            hid = int(hid)
        target_ids = list(range(hid, hid-n_results, -1))
        target_ids = [str(x) for x in target_ids if x > 0]
        output = []
        results = self.history.get(ids=target_ids, include=['documents', 'metadatas'])
        for tid in target_ids:
            if tid in results["ids"]:
                pos = results["ids"].index(tid)
                entry = results["metadatas"][pos]
                entry["content"] = results["documents"][pos]
                entry["id"] = tid
                output.append(entry)
        return output

    def get_history_from_last_batch(self):
        """
        Returns the history of the chat assistant from the last batch

        :return: The history
        :rtype: list
        """
        tid = self.get_current_id()
        results = self.history.get(ids=str(tid), include=['documents', 'metadatas'])
        if str(tid) in results["ids"]:
            pos = results["ids"].index(str(tid))
            entry = results["metadatas"][pos]
            entry["content"] = results["documents"][pos]
            entry["id"] = tid
            bid = entry["batch_id"]
            return self.get_batches([bid])
        return None

    def get_batch_before(self, bid):
        """
        Returns the history of the chat assistant from the last batch

        :return: The history
        :rtype: list
        """
        current_batch = self.get_batches([bid])
        lowest_id = None
        for entry in current_batch:
            tid = int(entry["id"])
            if lowest_id is None:
                lowest_id = tid
            if tid < lowest_id:
                lowest_id = tid
        in_next_batch = lowest_id - 1
        if in_next_batch < 0:
            return None
        result = self.get_history_from_id_and_earlier(in_next_batch, 1)
        if result is None:
            return None
        if len(result) == 0:
            return None
        next_batch_id = result[0]["batch_id"]
        return self.get_batches([next_batch_id])

    def get_batches(self, bids):
        """
        Returns the history of the chat assistant from the last batch

        :return: The history
        :rtype: list
        """
        output = []
        for bid in bids:
            results = self.history.get(include=['documents', 'metadatas'], where={"batch_id": str(bid)})
            ids = results["ids"]
            ids.sort()
            for tid in ids:
                pos = results["ids"].index(str(tid))
                entry = results["metadatas"][pos]
                entry["content"] = results["documents"][pos]
                entry["id"] = tid
                output.append(entry)
        time_stamp = []
        if self.time_injection:
            if len(output) > 0:
                stamp = 'Conversation below on ' + convert_utc_to_local(output[0]['utc_time']) + ': '
                time_stamp = output[0].copy()
                time_stamp['content'] = stamp
                time_stamp['role'] = 'system'
                time_stamp = [time_stamp]
        return time_stamp + output

    def get_summary_from_id_and_earlier(self, id=None, n_results=10):
        """
        Returns the summary of the chat assistant from a given id and earlier

        :param id: The id to start from
        :type id: int
        :param n_results: The number of results to return
        :type n_results: int
        :return: The history
        :rtype: list
        """
        if id is None:
            id = int(self.get_current_id(summary=True))
        else:
            id = int(id)
        target_ids = list(range(id, id-n_results, -1))
        target_ids = [str(x) for x in target_ids if x > 0]
        output = []
        results = self.summaries.get(ids=target_ids, include=['documents', 'metadatas'])
        for tid in target_ids:
            if tid in results["ids"]:
                pos = results["ids"].index(str(tid))
                entry = results["metadatas"][pos]
                entry["content"] = results["documents"][pos]
                entry["id"] = tid
                output.append(entry)
            if results["ids"].count(str(tid)) > 1:
                warnings.warn("Chat assistant: More than one summary found for id {}".format(id))
        return output

    def truncate_input_context(self, context):
        """
        Truncate the input context to the maximum number of tokens.

        :param context: The context to truncate.
        :type context: list
        :return: The truncated context.
        :rtype: list
        """
        while self.count_tokens_text(json.dumps(context)) > self.max_tokens:
            context.pop(1)
        return context



# __main__.py
from .config import set_openai_key, get_openai_key
import argparse
import os
import threading
import time
parser = argparse.ArgumentParser(description='Jarvis Conversationalist CLI')
parser.add_argument('--user', type=str, help='The username for the OpenAI API')
parser.add_argument('--key', type=str, help='The key for the OpenAI API')
parser.add_argument('--reset', action='store_true', help='Reset the saved username and key')
parser.add_argument('--verbose', action='store_true', help='Set the logging level to INFO')
parser.add_argument('--wipe_all_jarvis_memory', action='store_true', help='Reset Jarvis to factory defaults')
args = parser.parse_args()

os.environ['TOKENIZERS_PARALLELISM'] = "false"
if get_openai_key() is None:
    if args.key:
        set_openai_key(args.key)
        os.environ["OPENAI_API_KEY"] = get_openai_key()
    else:
        if not args.wipe_all_jarvis_memory:
            print("Please set your OpenAI API key using the --key argument once to cache your key.")
else:
    os.environ["OPENAI_API_KEY"] = get_openai_key()


from .config import load_key, load_config, save_config, set_openai_key, set_user, get_openai_key
from .logger_config import change_logging_level, get_log_folder_path

if args.wipe_all_jarvis_memory:
    import shutil

    log_path = get_log_folder_path()
    if os.path.exists(log_path):
        shutil.rmtree(log_path)
    db_path = os.path.join(os.path.expanduser('~'), 'Documents', "Jarvis DB")
    if os.path.exists(db_path):
        shutil.rmtree(db_path)
    quit(0)

lock_known_speakers = os.path.join(get_log_folder_path(), "known_speakers_db", "localaudiodb.lock")
lock_unknown_speakers = os.path.join(get_log_folder_path(), "unknown_speakers_db", "localaudiodb.lock")
if os.path.exists(lock_known_speakers):
    os.remove(lock_known_speakers)
if os.path.exists(lock_unknown_speakers):
    os.remove(lock_unknown_speakers)

from .conversationalist import converse
import warnings

warnings.filterwarnings("ignore", message="DeprecationWarning: 'audioop' is deprecated and "
                                          "slated for removal in Python 3.13")


def main():

    key = load_key()
    config = load_config(key)

    if args.verbose:
        change_logging_level("INFO")
    else:
        change_logging_level("ERROR")

    if args.reset:
        config = {}
    else:
        if args.user:
            config['user'] = args.user
        if args.key:
            config['key'] = args.key

    save_config(config, key)

    set_openai_key(config.get('key', None))
    set_user(config.get('user', 'User'))

    if get_openai_key() is None:
        print("Please set your OpenAI API key using the --key argument once to cache your key.")
        return
    print("Please wait while Jarvis boots...", end='\r')
    interrupt_event = threading.Event()
    start_event = threading.Event()
    stop_event = threading.Event()
    conversation_thread = threading.Thread(target=converse, args=(60, interrupt_event, start_event, stop_event),)
    conversation_thread.start()

    hello = "\rSay my name, 'Jarvis', to get my attention... Click \"Enter\" to interrupt me. " \
            "Click \"Esc\" then \"Enter\" to turn me off.\033[K"
    LINE_UP = '\033[1A'
    LINE_CLEAR = '\x1b[2K'
    try:
        while True:
            if conversation_thread.is_alive():
                if not start_event.is_set():
                    print("\033[KI am booting, be quiet - "
                          "I'm sampling background noise levels in your space...\033[K", end='\r')
                    start_event.wait(timeout=1)
                else:
                    user_input = input(hello)
                    if user_input.strip() == '' and not interrupt_event.is_set():
                        interrupt_event.set()
                        # Clear the line and then print "Interrupted."
                        print(end=LINE_CLEAR)
                        print(LINE_UP, end=LINE_CLEAR)
                        print("\033[KSorry for going on and on.\033[K", end='\r')
                        reset_time = time.time()
                    elif user_input == '\x1b':
                        # Clear the line and then print "Quitting Jarvis"
                        print("\r\033[KI am starting my shutdown process...\033[K", end='\r')
                        stop_event.set()
                        while conversation_thread.is_alive():
                            threading.Event().wait(timeout=10)
                        conversation_thread.join(timeout=10)
                        return
            else:
                raise Exception("Conversation process is not alive!")

    except KeyboardInterrupt:
        print()
        print("Shutting down ASAP! Please Wait...")
        stop_event.set()
        while conversation_thread.is_alive():
            print("Shutting down...")
            threading.Event().wait(timeout=10)
        conversation_thread.join(timeout=10)
        return


if __name__ == "__main__":
    main()



import json
import certifi
import os
from openai import OpenAI, _utils

client = OpenAI()
_utils._logs.logger.setLevel("CRITICAL")

os.environ["REQUESTS_CA_BUNDLE"] = certifi.where()
os.environ["SSL_CERT_FILE"] = certifi.where()

name = "Jarvis"


def check_for_directed_at_me(transcript, n=1):
    functions = [
        {
            "name": "configure_response",
            "description": "Configures the appropriate response to the user's input.",
            "parameters": {
                "type": "object",
                "properties": {
                    "probability": {
                        "type": "number",
                        "description": "from 0 to 100, the likelihood that the user was speaking directly to"
                                       " and not about "+name+". Remember that the user may be speaking to someone "
                                       "else and that should factor into the probability that the user is speaking "
                                       "to "+name+" as opposed to "
                                                  "about "+name+". For example, if the user says '"+name+", "
                                       "please turn on the lights', the probability should be high that the user is "
                                       "speaking to "+name+". However, if the user says 'I was talking to "+name+" "
                                       "and he turned on the lights', the probability should be low (63) that the user "
                                       "is speaking to "+name+". Or, if the user says 'John, please turn on the lights "
                                       "for me', the probability should be low (15) that the user is speaking"
                                                              " to "+name+"."
                    }
                },
                "required": ["probability"],
            },
        }
    ]
    system_message = "You are seeing a live transcription of what is being said in a room." \
                     " There are many reasons why names might be mentioned. There may also be multiple" \
                     " people in the room or people on the phone. It is your job to determine if the user is speaking" \
                     " to " + name + " directly."

    response = client.chat.completions.create(model="gpt-3.5-turbo",
                                              messages=[{"role": "system", "content": system_message},
                                                        {"role": "user", "content": "\n".join(transcript)}],
                                              functions=functions,
                                              n=n,
                                              function_call={"name": "configure_response"})
    probabilities = []
    for result in response.choices:
        result = result.message.function_call
        if result:
            probabilities.append(json.loads(result.arguments)['probability']/100)
    return probabilities


def check_for_completion(transcript, n=1):
    """
    What is the likelihood that the user is done speaking?

    :param transcript: an array of strings representing the user's speech
    :param n: the number of responses to generate
    :return: list of likelihoods that the user is done speaking
    """
    functions = [
        {
            "name": "configure_response",
            "description": "Configures the appropriate response to the user's input.",
            "parameters": {
                "type": "object",
                "properties": {
                    "probability": {
                        "type": "number",
                        "description": "A value from 0 to 100, the probability that the user is done speaking based on "
                                       "ALL of the following core principles: "
                                       "The user has completed their full thought. The user has "
                                       "completed their sentence. The user has included all components of their "
                                       "response that they intended to include at the beginning of their statement."
                                       "If the user has not completed their thought, sentence, or included all parts"
                                       " of their response, the probability should be low (below 10). "
                                       "If the user ended their input without punctuation, and 'and' or 'but', the "
                                       "probability should be low (below 10)."
                                       "If the user has completed some of their thought, sentence, or included some "
                                       "parts of their response, the probability should be medium (between 10 and 55)."
                                       "If the user has completed their thought, sentence, and included all parts of "
                                       "their response, the probability should be high (above 72)."
                    }
                },
                "required": ["probability"],
            },
        }
    ]
    system_message = "You are seeing a live transcription of what is being said in a room. It is your job to determine"
    " if the user is done speaking by analyzing the text below and seeing if the user has completed their thought."

    response = client.chat.completions.create(model="gpt-4",
                                              messages=[{"role": "system", "content": system_message},
                                                        {"role": "user", "content": "\n".join(transcript)}],
                                              functions=functions,
                                              n=n,
                                              function_call={"name": "configure_response"})
    probabilities = []
    for result in response.choices:
        result = result.message.function_call
        if result:
            probabilities.append(json.loads(result.arguments)["probability"]/100)
    return probabilities


def extract_query(transcript):
    """
    Extracts the query from the user's speech.

    :param transcript: an array of strings representing the user's speech
    :return: the query
    """
    functions = [
        {
            "name": "configure_response",
            "description": "Configures the appropriate response to the user's input.",
            "parameters": {
                "type": "object",
                "properties": {
                    "query": {
                        "type": "string",
                        "description": "The query that the user is asking. For example, if the user says 'What is the "
                                       "weather like today "+name+"?', the query should be "
                                       "'What is the weather like today, "+name+"?'. Or if the user says '"+name+", "
                                       "please turn on the lights', the query should be "
                                       "'"+name+", Please turn on the lights'."
                    }
                },
                "required": ["query"],
            },
        }
    ]
    system_message = "You are seeing a live transcription of what is being said in a room. It is your job to " \
                     "determine the query that the user is asking by analyzing the text below and extracting word " \
                     "for word the section of the transcript that is the query, include who is asking it using " \
                     "the bracket format like '[Unknown Speaker X]: ' or '[Jane]: ' or '[John Doe]: ' etc. " \
                     "Ignore the rest of the unrelated transcript. Keep in mind that sometimes context from " \
                     "earlier parts of the conversation are critical to understanding a query - make sure to include" \
                     "all the context needed to complete the query well. " \
                     "There may be multiple people in the room or people " \
                     "on the phone. It is your job to determine which part of the transcript is the query meant " \
                     "for " + name + ". The query should be a question or a command or a statement directed at or" \
                                     "highly related to " + name + ".  It is ok if there are multiple parts of the " \
                                                                   "query, or if multiple people appear to be asking " \
                                                                   "questions to" + name + ", it is ok to include " \
                                                                                           "all of those subsections " \
                                                                                           "in the query - just make " \
                                                                                           "sure to include the " \
                                                                                           "speaker annotation for " \
                                                                                           "each subsection."

    response = client.chat.completions.create(model="gpt-4",
                                              messages=[{"role": "system", "content": system_message},
                                                        {"role": "user", "content": "\n".join(transcript)}],
                                              functions=functions,
                                              function_call={"name": "configure_response"})
    result = response.choices[0].message.function_call
    if result:
        return json.loads(result.arguments)["query"]
    return


def simple_stream_response(query, context=None):
    if context is None:
        context = []
    return client.chat.completions.create(model="gpt-4",
                                          messages=context+[{"role": "user", "content": query}],
                                          stream=True)



import json

from .internet_helper import create_internet_context
from .internet_helper import search
from .weather_functions import get_weather
from .google_tools import simple_google_search
from .basic_tools import write_to_file, open_webpage


def deep_long_duration_internet_research_documentation():
    """
    Returns json documentation for the internet context function.

    :return: The json documentation for the internet context function.
    """
    schema = {"type": "function",
              "function": {
                "name": "deep_long_duration_internet_research",
                "description": "Searches the internet for a query and returns a detailed summary of the top "
                               "most relevant results.",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "query": {
                            "type": "string",
                            "description": "The query to search for.",
                        },
                    },
                    "required": ["query"],
                },
              }
              }
    examples = 'Examples:\n {"function_name": "search_helper", "parameters": {"query": "What is the current state' \
               'of the Israeli Palestinian Conflict"}}\n' \
               '{"function_name": "search_helper", "parameters": {"query": "What impactful new climate change ' \
               'research was released this month"}}\n'
    return schema, examples


def deep_long_duration_internet_research(query=None):
    return create_internet_context(query)


def search_weather(query=None):
    return get_weather(query)


def search_weather_documentation():
    """
    Returns json documentation for the get weather function.

    :return: The json documentation for the get weather function.
    """
    schema = {"type": "function",
              "function": {
                    "name": "search_weather",
                    "description": "Gets the weather for a location.",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "query": {
                                "type": "string",
                                "description": "The location to get the weather for.",
                            },
                        },
                        "required": ["query"],
                    },
              }
              }
    examples = 'Examples:\n {"function_name": "search_weather", "parameters": {"query": "Paris France"}}\n ' \
               '{"function_name": "search_weather", "parameters": {"query": "Baltimore, MD"}}\n' \
               '{"function_name": "search_weather", "parameters": {"query": "New York City"}}\n'
    return schema, examples


def quick_google_answer(query=None):
    return simple_google_search(query)


def quick_google_answer_documentation():
    schema = {"type": "function",
                "function": {
                    "name": "quick_google_answer",
                    "description": "Searches google for query and returns simple processed answers"
                                   " if Google returns one",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "query": {
                                "type": "string",
                                "description": "The query to search for.",
                            },
                        },
                        "required": ["query"],
                    },
                }
                }
    examples = 'Examples:\n {"function_name": "quick_google_answer", "parameters": {"query": "What is population of ' \
               'france?"}}\n {"function_name": "quick_google_answer", "parameters": {"query": "What is the capital ' \
               'of France?"}}\n {"function_name": "quick_google_answer", "parameters": {"query": "What is the ' \
               'value of pi?"}}\n'
    return schema, examples


def get_top_google_results(query=None):
    return json.dumps(search(query, num_results=10, advanced=True)['items'])


def get_top_google_results_documentation():
    schema = {"type": "function",
                "function": {
                    "name": "get_top_google_results",
                    "description": "Searches google for query and returns the top 10 results",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "query": {
                                "type": "string",
                                "description": "The query to search for.",
                            },
                        },
                        "required": ["query"],
                    },
                }
                }
    examples = 'Examples:\n {"function_name": "get_top_google_results", "parameters": {"query": "Youtube video of how ' \
               'to make a cake"}}\n {"function_name": "get_top_google_results", "parameters": {"query": "Recipes to ' \
               'make a cake"}}'
    return schema, examples


def write_to_a_file(name=None, text=None):
    return write_to_file(name, text)


def write_to_a_file_documentation():
    schema = {"type": "function",
                "function": {
                    "name": "write_to_a_file",
                    "description": "Writes text to a file with the given name",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "name": {
                                "type": "string",
                                "description": "The name of the file to write to.",
                            },
                            "text": {
                                "type": "string",
                                "description": "The text to write to the file.",
                            },
                        },
                        "required": ["name", "text"],
                    },
                }
                }
    examples = 'Examples:\n {"function_name": "write_to_a_file", "parameters": {"name": "test.txt", "text": "This ' \
                  'is a test"}}\n {"function_name": "write_to_a_file", "parameters": {"name": "hello.json", "text": ' \
                    '"{\\"hello\\": \\"world\\"}"}}\n {"function_name": "write_to_a_file", "parameters": {"name": ' \
                    '"test.md", "text": "# This is a test"}}\n'
    return schema, examples


def open_url_for_user(url=None):
    return open_webpage(url)


def open_url_for_user_documentation():
    schema = {"type": "function",
                "function": {
                    "name": "open_url_for_user",
                    "description": "Opens a url in the user's default web browser.",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "url": {
                                "type": "string",
                                "description": "The url to open.",
                            },
                        },
                        "required": ["url"],
                    },
                }
                }
    examples = 'Examples:\n {"function_name": "open_url_for_user", "parameters": {"url": "https://www.google.com"}}\n' \
               '{"function_name": "open_url_for_user", "parameters": {"url": "https://www.youtube.com"}}\n'
    return schema, examples


def get_function_info():
    return {"deep_long_duration_internet_research": {"function": deep_long_duration_internet_research,
                                                     "schema": deep_long_duration_internet_research_documentation()[0],
                                                     "examples": deep_long_duration_internet_research_documentation()[1]},
            "search_weather": {"function": search_weather,
                            "schema": search_weather_documentation()[0],
                            "examples": search_weather_documentation()[1]},
            "quick_google_answer": {"function": quick_google_answer,
                                    "schema": quick_google_answer_documentation()[0],
                                    "examples": quick_google_answer_documentation()[1]},
            "get_top_google_results": {"function": get_top_google_results,
                                        "schema": get_top_google_results_documentation()[0],
                                        "examples": get_top_google_results_documentation()[1]},
            "write_to_a_file": {"function": write_to_a_file,
                                "schema": write_to_a_file_documentation()[0],
                                "examples": write_to_a_file_documentation()[1]},
            "open_url_for_user": {"function": open_url_for_user,
                                    "schema": open_url_for_user_documentation()[0],
                                    "examples": open_url_for_user_documentation()[1]},
            }


info = get_function_info()
function_list = []
function_examples = []
for val in info.values():
    function_list.append(val["schema"])
    function_examples.append(val["examples"])


def get_function_list():
    return function_list


def get_function_examples():
    return function_examples


def get_system_appendix():
    base_message = "\n\nFunction Calling:\nDon't make assumptions about what values to plug into functions. " \
                   "Ask for clarification if a user request is ambiguous. Only use the functions in your " \
                   "function list. Make sure to always include all required parameters and to use the correct " \
                   "data types in the correct JSON format.\n"
    return base_message + "\n".join(function_examples)







import multiprocessing
import queue
import threading
import re
import wave
import io
import warnings
import spacy
import atexit
from queue import Queue
from numpy import frombuffer, int16
from typing import Iterator, Dict, Tuple, Optional
from .text_speech import text_to_speech, TextToSpeechError

warnings.filterwarnings("ignore", message=".*The rule-based lemmatizer did not find POS annotation.*")

try:
    nlp = spacy.load('en_core_web_sm', disable=["tagger", "parser", "ner"])
except OSError:
    print("Model 'en_core_web_sm' not found. Downloading...")
    spacy.cli.download('en_core_web_sm')
    nlp = spacy.load('en_core_web_sm', disable=["tagger", "parser", "ner"])
nlp.add_pipe("sentencizer")

CHUNK = 8196
CHANNELS = 1
RATE = 16000

rt_text_queue_global = None


class SpeechStreamer:
    """
    A class for streaming text-to-speech audio.

    :param stop_other_audio: A threading.Event() to stop any currently playing audio.
    :type stop_other_audio: threading.Event(), optional
    :param skip: A threading.Event() to skip the current audio stream.
    :type skip: threading.Event(), optional
    :param rt_queue: A queue to send real-time transcription data.
    :type rt_queue: queue.Queue(), optional
    """

    def __init__(self, stop_other_audio: threading.Event = None,
                 skip: threading.Event = None, rt_queue: queue.Queue = None) -> None:
        """
        Initializes the SpeechStreamer class with default values.

        The class instance is created with a queue for handling the text-to-speech
        audio, and the corresponding thread is started to play the audio. The PyAudio
        module is also used to initialize the class instance with parameters to control
        the audio format and output.

        :param stop_other_audio: A threading.Event() to stop any currently playing audio.
        :type stop_other_audio: threading.Event(), optional
        :param skip: A threading.Event() to skip the current audio stream.
        :type skip: threading.Event(), optional
        :param rt_queue: A queue to send real-time transcription data.
        :type rt_queue: queue.Queue(), optional
        """
        self.queue = queue.Queue()
        if rt_queue is None:
            rt_queue = multiprocessing.Queue()
        if stop_other_audio is None:
            stop_other_audio = multiprocessing.Event()
        if skip is None:
            skip = multiprocessing.Event()
        self.rt_queue = rt_queue
        self.playing = False
        self.thread = threading.Thread(target=self._play_audio, args=(stop_other_audio, skip))
        self.thread.daemon = True
        self.thread.start()
        atexit.register(self.stop)
        self.stop_event = threading.Event()
        self.skip = skip
        self.audio_count = 0
        self.lock = threading.Lock()
        self.done = False
        self.last_events = {}

    def _play_audio(self, stop_other_audio: threading.Event = None,
                    skip: threading.Event = None) -> None:
        """
        Plays the audio stream generated from the text-to-speech data.

        The audio stream is played by iterating over the audio buffer chunk by chunk,
        and writing each chunk to the PyAudio instance. The class instance is also
        initialized with a threading.Lock() to prevent conflicts when accessing
        the audio_count variable.

        :param stop_other_audio: A threading.Event() to stop any currently playing audio.
        :type stop_other_audio: threading.Event(), optional
        :param skip: A threading.Event() to skip the current audio stream.
        :type skip: threading.Event(), optional
        """
        import sounddevice as sd
        while True:
            generator, sample_rate = self.queue.get()
            if generator is None:
                continue

            if stop_other_audio:
                stop_other_audio.set()

            if not self.playing:
                self.playing = True

            chunk_played = False
            with sd.OutputStream(samplerate=sample_rate, channels=CHANNELS, dtype='int16') as stream:
                for chunk in generator():
                    if skip and skip.is_set():
                        self.stop()
                        return
                    stream.write(chunk)
                    chunk_played = True

            self.playing = False

            if chunk_played:
                with self.lock:
                    self.audio_count -= 1
                    if self.audio_count == 0 and self.done:
                        self.stop_event.set()

    def queue_text(self, text: str, delay: float = 0, model: str = "gpt-4") -> None:
        """
        Queues the text data for text-to-speech processing.

        The text data is queued and passed to the _process_text_to_speech() function to be
        converted to audio. A threading.Lock() is used to prevent conflicts when accessing
        the audio_count variable.

        :param text: The text data to be converted to speech.
        :type text: str
        :param delay: The delay time in seconds before playing the audio.
        :type delay: float, optional
        :param model: The text-to-speech model to be used for conversion.
        :type model: str, optional
        """
        with self.lock:
            self.audio_count += 1
            current_event = threading.Event()
            self.last_events[self.audio_count] = current_event
            last_event = None
            if self.audio_count > 1:
                last_event = self.last_events[self.audio_count-1]
        tts_thread = threading.Thread(target=self._process_text_to_speech,
                                      args=(text, delay, model, self.rt_queue,
                                            current_event, last_event))
        tts_thread.daemon = True
        tts_thread.start()

    def stop(self) -> None:
        """
        Stops the audio stream and PyAudio instance.

        The function stops the audio stream and PyAudio instance by closing the audio
        stream and terminating the PyAudio instance. A threading.Event() is used to
        ensure that the audio stream is stopped before closing the stream.

        """
        self.done = True
        if self.skip is None:
            self.stop_event.wait()
        else:
            while self.stop_event.is_set() is False and self.skip.is_set() is False:
                self.skip.wait(timeout=1)

    def _process_text_to_speech(self, text: str, delay: float, model: str,
                                rt_text: queue.Queue, current: threading.Event,
                                last: threading.Event) -> None:
        """
        Processes the text data into audio format.

        The text data is passed to the text_to_speech() function for processing
        and conversion into audio format. The audio data is then converted to a numpy array
        and passed to the queue to be played as audio. The real-time transcription data
        is also added to the queue.

        :param text: The text data to be converted to speech.
        :type text: str
        :param delay: The delay time in seconds before playing the audio.
        :type delay: float
        :param model: The text-to-speech model to be used for conversion.
        :type model: str
        :param rt_text: A queue to send real-time transcription data.
        :type rt_text: queue.Queue()
        :param current: An event to let others know this was added
        :type current: threading.event()
        :param last: An event to make sure the last text is added
        :type last: threading.event()
        """
        try:
            byte_data = text_to_speech(text, stream=True, model=model)
        except TextToSpeechError:
            return
        self.skip.wait(timeout=delay)
        wav_io = io.BytesIO(byte_data)
        with wave.open(wav_io, 'rb') as wav_file:
            n_channels, sample_width, frame_rate, n_frames = wav_file.getparams()[:4]
            audio_data = wav_file.readframes(n_frames)

        # Convert audio data to numpy array
        np_audio_data = frombuffer(audio_data, dtype=int16)

        def generator():
            for i in range(0, len(np_audio_data), CHUNK):
                yield np_audio_data[i:i + CHUNK]
        rt_text.put({"role": "assistant", "content": text, "model": model})
        if last:
            last.wait()
        self.queue.put((generator, frame_rate))
        current.set()


def stream_audio_response(streaming_text: Iterator[Dict], stop_audio_event: Optional[threading.Event] = None,
                          skip: Optional[threading.Event] = None) -> Tuple[str, str, dict]:
    """
    Streams the audio response.

    The function receives a streaming_text object containing the text data
    that needs to be converted to speech. The text data is processed and converted
    to speech using the SpeechStreamer class. The real-time transcription data is
    passed to the global rt_text_queue_global variable.

    :param streaming_text: An iterator containing the streaming text data.
    :type streaming_text: Iterator[Dict]
    :param stop_audio_event: An event object to stop the audio stream, defaults to None.
    :type stop_audio_event: threading.Event, optional
    :param skip: An event object to skip the audio stream, defaults to None.
    :type skip: threading.Event, optional
    :return: A tuple containing the output text and the finish reason.
    :rtype: Tuple[str, str, dict]
    """
    global rt_text_queue_global
    if rt_text_queue_global is None:
        rt_text_queue_global = multiprocessing.Queue()
    if stop_audio_event is None:
        stop_audio_event = multiprocessing.Event()
    if skip is None:
        skip = multiprocessing.Event()
    speech_stream = SpeechStreamer(stop_other_audio=stop_audio_event, skip=skip, rt_queue=rt_text_queue_global)
    buffer = ""
    output = ""
    tool_calls = {}
    resp = None
    delay = 0.0
    model = None
    for resp in streaming_text:
        if skip:
            if skip.is_set():
                break
        if resp.choices:
            if model is None:
                model = resp.model
            if resp.choices[0].delta.content:
                text = resp.choices[0].delta.content
                if text is not None:
                    buffer += text
                    output += text
                    doc = nlp(buffer)
                    sentences = list(doc.sents)

                    if len(sentences) > 1:
                        merged_sentences = []
                        i = 0
                        while i < len(sentences) - 1:
                            current_sentence = sentences[i].text.strip()
                            next_sentence = sentences[i + 1].text.strip()

                            if len(current_sentence) < 50:
                                current_sentence += " " + next_sentence
                                i += 2
                            else:
                                i += 1
                            merged_sentences.append(current_sentence)

                        if i == len(sentences) - 1:
                            merged_sentences.append(sentences[-1].text.strip())

                        for sentence in merged_sentences[:-1]:
                            if len(re.sub('[^a-z|A-Z|0-9]', '', sentence)) > 1:
                                speech_stream.queue_text(sentence, delay=delay, model=model)
                                delay = 0

                        # Keep the last part (which may be an incomplete sentence) in the buffer
                        buffer = merged_sentences[-1]
            if resp.choices[0].delta.tool_calls:
                for tool in resp.choices[0].delta.tool_calls:
                    if tool.function:
                        if tool.function.name:
                            tool_calls[tool.index] = {"name": tool.function.name}
                        elif tool.function.arguments:
                            if "arguments" in tool_calls[tool.index]:
                                tool_calls[tool.index]["arguments"] += tool.function.arguments
                            else:
                                tool_calls[tool.index]["arguments"] = tool.function.arguments
                    else:
                        warnings.warn("Tool call does not contain a function.")

    if skip:
        if skip.is_set():
            return "Sorry.", "null"
    if resp:
        reason = resp.choices[0].finish_reason
    else:
        reason = "null"
    if reason != "stop":
        if reason == "length":
            buffer += "... I'm sorry, I have been going on and on haven't I?"
            output += "... I'm sorry, I have been going on and on haven't I?"
        if reason == "null":
            buffer += " I'm so sorry I got overwhelmed, can you put that more simply?"
            output += " I'm so sorry I got overwhelmed, can you put that more simply?"
        if reason == "content_filter":
            buffer += " I am so sorry, but if I responded to that I would have been forced to say something naughty."
            output += " I am so sorry, but if I responded to that I would have been forced to say something naughty."
        if reason == "function_call":
            buffer += " Processing..."
            output += " Processing..."
    if len(re.sub('[^a-z|A-Z|0-9]', '', buffer)) > 1:
        speech_stream.queue_text(buffer)
    else:
        if buffer == output:
            speech_stream.queue_text("Processing...")
    speech_stream.stop()
    if len(tool_calls) == 0:
        tool_calls = None
    return output, reason, tool_calls


def set_rt_text_queue(rt_text_queue: Queue) -> None:
    """
    Set the global rt_text_queue_global variable to the given rt_text_queue.

    :param rt_text_queue: A queue containing the real-time text data.
    :type rt_text_queue: Queue
    """
    global rt_text_queue_global
    rt_text_queue_global = rt_text_queue



import logging
import multiprocessing
multiprocessing.set_start_method('spawn')
import threading
import unittest
import os
from datetime import datetime

from src.jarvis_conversationalist.config import get_openai_key
if get_openai_key() is not None:
    os.environ["OPENAI_API_KEY"] = get_openai_key()
from src.jarvis_conversationalist.logger_config import get_logger
from src.jarvis_conversationalist.conversationalist import process_assistant_response, get_core_path, converse
from src.jarvis_conversationalist.audio_player import play_audio_file
from src.jarvis_conversationalist.openai_functions.functions import get_function_info
from src.jarvis_conversationalist.audio_listener import audio_capture_process, listen_to_user, prep_mic


class TestJarvisConversationalist(unittest.TestCase):

    def test_get_logger(self):
        # Define a test case
        logger = get_logger()
        # Assert that the function returns the expected result
        self.assertIsNotNone(logger)

    def test_play_audio_file(self):
        # Define a test case
        file_path = get_core_path() + "/tone_two.wav"
        blocking = True
        play_audio_file(file_path, blocking)
        self.assertIsNotNone(blocking)

    # def test_listen_to_user(self):
    #     # Define a test case
    #     level = prep_mic()
    #     audio_data = listen_to_user(level)
    #     self.assertIsNotNone(audio_data)
    #
    # def test_audio_process(self):
    #     audio_queue = multiprocessing.Queue()
    #     speaking = multiprocessing.Event()
    #     multiprocessing_stop_event = multiprocessing.Event()
    #     capture_process = multiprocessing.Process(target=audio_capture_process,
    #                                               args=(audio_queue, speaking, multiprocessing_stop_event), )
    #     capture_process.start()
    #     self.assertTrue(capture_process.is_alive())
    #     # wait for audio queue to be populated
    #     audio_queue.get(timeout=60)

    def test_converse(self):
        # Define a test case
        timeout = 60
        interrupt_event = threading.Event()
        start_event = threading.Event()
        stop_event = threading.Event()
        conversation_thread = threading.Thread(target=converse, args=(timeout,
                                                                      interrupt_event,
                                                                      start_event,
                                                                      stop_event),)
        conversation_thread.start()
        self.assertIsNotNone(conversation_thread)
        get_logger().info(str(start_event.is_set())+" "+str(datetime.now()))
        start_event.wait(timeout=140)
        threading.Event().wait(timeout=5)
        get_logger().info(str(start_event.is_set()) + " " + str(datetime.now()))
        stop_event.set()
        threading.Event().wait(timeout=25)
        get_logger().info(threading.enumerate())
        conversation_thread.join(timeout=60)
        get_logger().info(threading.enumerate())
        if not conversation_thread.is_alive():
            closed = True
        else:
            closed = False
        self.assertTrue(closed)

    def test_process_assistant_response(self):
        # Define a test case
        beeps_stop_event = multiprocessing.Event()
        interrupt_event = threading.Event()
        query = "What's the weather in Baltimore?"
        context = process_assistant_response(query, beeps_stop_event, interrupt_event)
        # Assert that the function returns the expected result
        self.assertIsNotNone(context)



import logging
import os
from logging.handlers import RotatingFileHandler


user_home = os.path.expanduser("~")
logs_dir = os.path.join(user_home, "Jarvis Logs")
if not os.path.exists(logs_dir):
    os.mkdir(logs_dir)

level_path = os.path.join(logs_dir, "level.txt")
logger_path = os.path.join(logs_dir, "jarvis_process.log")
logger_stream = logging.StreamHandler()
logger_stream.setFormatter(logging.Formatter('\033[K%(asctime)s [%(levelname)s] - %(message)s'))
rotating_handler = RotatingFileHandler(logger_path, maxBytes=1024*10, backupCount=3)
# Set up a logger
logging.basicConfig(level=logging.INFO,
                    format='%(asctime)s [%(levelname)s] - %(message)s',
                    handlers=[logging.FileHandler(logger_path), rotating_handler])

if os.path.exists(level_path):
    with open(level_path, 'r') as f:
        wlevel = f.read()
    logging.getLogger().setLevel(wlevel)
    if wlevel != "ERROR":
        logging.getLogger().addHandler(logger_stream)


def get_logger():
    """
    Sets up the jarvis_process.log

    :return: The logger.
    :rtype: logging.Logger
    """
    return logging.getLogger(__name__)


def change_logging_level(level):
    """
    Change the logging level of the logger.

    :return: None
    :rtype: None
    """
    logging.getLogger().setLevel(level)
    with open(level_path, 'w') as f:
        f.write(level)
    if level != "ERROR":
        logging.getLogger().addHandler(logger_stream)

def get_log_folder_path():
    """
    Get the path to the log folder.

    :return: The path to the log folder.
    :rtype: str
    """
    return logs_dir



import multiprocessing
import queue
import threading
import time
import os
import re
from statistics import mean
import certifi
from numpy import average
import importlib.resources as pkg_resources

from .openai_utility_functions import check_for_directed_at_me, check_for_completion, extract_query
from .openai_interface import stream_response, resolve_response, use_tools, schedule_refresh_assistant
from .streaming_response_audio import stream_audio_response, set_rt_text_queue
from .audio_player import play_audio_file
from .audio_listener import audio_capture_process
from .audio_transcriber import audio_processing_thread

from .logger_config import get_logger
logger = get_logger()


try:
    core_path = str(pkg_resources.files('jarvis_conversationalist').joinpath('audio_files'))
except Exception as e:
    if os.getcwd().find('jarvis-conversationalist') != -1:
        core_path = os.path.join(os.getcwd()[:os.getcwd().find('jarvis-conversationalist')],
                                 'jarvis-conversationalist', 'src', 'jarvis_conversationalist', 'audio_files')
    else:
        raise Exception(f"Failed to find the audio files directory: {e}")

os.environ["REQUESTS_CA_BUNDLE"] = certifi.where()
os.environ["SSL_CERT_FILE"] = certifi.where()

wake_word = "jarvis"


def get_core_path():
    """
    Get the path to the core audio files.
    :return: The path to the core audio files.
    :rtype: str
    """
    return core_path


def process_assistant_response(query, beeps_stop_event, interrupt_event):
    """
    Processes the assistant's response to the user's query.
    :param query:
    :param beeps_stop_event:
    :param interrupt_event:
    :return: History of the conversation.
    :rtype: list
    """
    new_history = [{"content": query, "role": "user"}]
    current_stream = queue.Queue()
    set_rt_text_queue(current_stream)
    content, reason, tool_calls = stream_audio_response(stream_response(new_history),
                                                        stop_audio_event=beeps_stop_event,
                                                        skip=interrupt_event)
    if not interrupt_event.is_set():
        if tool_calls is None:
            new_history.append({"content": content, "role": "assistant"})
            logger.info("Response Content:"+" "+str(content))
        else:
            while tool_calls is not None and not interrupt_event.is_set():
                tool_stop_event = play_audio_file(core_path+"/searching.wav", loops=7, blocking=False)
                new_history += use_tools(tool_calls, content)
                tool_stop_event.set()
                content, reason, tool_calls = stream_audio_response(
                    stream_response(new_history, keep_last_history=True),
                    stop_audio_event=beeps_stop_event,
                    skip=interrupt_event)
            if tool_calls is None:
                new_history.append({"content": content, "role": "assistant"})
    if interrupt_event.is_set():
        logger.info("Interrupted")
        partial = ""
        while not current_stream.empty():
            partial += current_stream.get()["content"] + "\n"
        new_history.append({"content": partial, "role": "assistant"})
        new_history.append({"content": "Interupted by user.", "role": "system"})
    return new_history


def converse(memory, interrupt_event, start_event, stop_event):
    """
    Converse with the user.
    :param memory:
    :type memory: int
    :param interrupt_event: An event to indicate that the user has interrupted the assistant.
    :type interrupt_event: threading.Event
    :param start_event: An event to indicate that the assistant has started.
    :type start_event: threading.Event
    :param stop_event: An event to indicate that the assistant should stop.
    :type stop_event: threading.Event
    :return:
    """
    audio_queue = multiprocessing.Queue()
    text_queue = multiprocessing.Queue()
    multiprocessing_stop_event = multiprocessing.Event()

    speaking = multiprocessing.Event()
    capture_process = multiprocessing.Process(target=audio_capture_process,
                                              args=(audio_queue, speaking, multiprocessing_stop_event), )
    text_process = multiprocessing.Process(target=audio_processing_thread,
                                           args=(audio_queue, text_queue, speaking, multiprocessing_stop_event), )

    capture_process.start()
    text_process.start()

    # Rolling buffer to store text
    transcript = []
    timestamps = []
    schedule_refresh_assistant()
    logger.info("Waiting for text queue...")
    while text_queue.empty() and not stop_event.is_set():
        threading.Event().wait(1)
    logger.info("Starting...")
    while not text_queue.empty() and not stop_event.is_set():
        try:
            text_queue.get(timeout=5)
        except multiprocessing.queues.Empty:
            pass
    play_audio_file(core_path + "/tone_one.wav", blocking=False)
    start_event.set()
    delays = []
    while not stop_event.is_set():
        interrupt_event.clear()
        try:
            text, ts = text_queue.get(timeout=1)
            text = text.strip()
            if text == "Thank you for watching." or text == "Thanks for watching!"\
                    or text == "Thanks for watching." or text == "Thank you for watching!"\
                    or text == "Thanks for watching!" or text == "You":
                text = ""
            if bool(re.search('[a-zA-Z0-9]', text)):
                delays.append(time.time() - ts)
                avg_delay = average(delays)
                if len(delays) > 10:
                    delays.pop(0)
                logger.info("Average delay:" + " " + str(avg_delay) + " Raw text: " + text)
                transcript.append(text)
                timestamps.append(time.time())
        except queue.Empty:
            text = ""
            # wait for 1 second
            threading.Event().wait(1)
        if text != "":
            # Clean up old entries
            current_time = time.time()
            while transcript and timestamps[0] < current_time - memory:
                transcript.pop(0)
                timestamps.pop(0)

            # Check for "jarvis" and print buffer if found
            if wake_word in text.lower():
                logger.info(" - - Checking if what has been said was directed at me...")
                logger.info("\n".join(transcript))
                directed_at_results = check_for_directed_at_me(transcript)
                directed_at_results = [x for x in directed_at_results if x <= 1.0]
                if len(directed_at_results) == 0:
                    probably_at_me = False
                else:
                    target_intended = 0.7
                    probably_at_me = average(directed_at_results) > target_intended
                logger.info("Probability at me: " + str(directed_at_results))
                if probably_at_me:
                    completion_results = check_for_completion(transcript)
                    logger.info("Probability of completion:" + str(completion_results))
                    completion_results = [x for x in completion_results if x <= 1.0]
                    target_completion = 0.7
                    if len(completion_results) == 0:
                        completed = False
                    else:
                        completed = mean(completion_results) > target_completion
                    if not completed:
                        additions = 0
                        max_time = 25
                        max_additions = 3
                        current_time = time.time()
                        while time.time() - current_time < max_time and additions < max_additions:
                            if not text_queue.empty():
                                text, ts = text_queue.get()
                                if bool(re.search('[a-zA-Z0-9]', text)):
                                    logger.info(" - - - Adding to transcript: " + text)
                                    transcript.append(text)
                                    timestamps.append(time.time())
                                    additions += 1
                                    completion_results = check_for_completion(transcript)
                                    logger.info("Still checking probability of completion:"+" "+str(completion_results))
                                    completion_results = [x for x in completion_results if x <= 1.0]
                                    if len(completion_results) == 0:
                                        completed = False
                                    else:
                                        completed = mean(completion_results) > target_completion
                                    if completed:
                                        additions = max_additions
                            else:
                                threading.Event().wait(0.3)
                    speaking.set()
                    beeps_stop_event = play_audio_file(core_path+"/beeps.wav", loops=7, blocking=False)
                    extracted_query = extract_query(transcript)
                    logger.info("Query extracted: " + extracted_query)
                    new_history = None
                    if not interrupt_event.is_set():
                        #try:
                            new_history = process_assistant_response(extracted_query, beeps_stop_event, interrupt_event)
                        # except Exception as e:
                        #     logger.error(e)
                        #     play_audio_file(core_path + "/major_error.wav", blocking=False)
                        #     new_history = [{"content": extracted_query, "role": "user"},
                        #                    {"content": "I'm sorry, I'm having so issues with my circuits.",
                        #                     "role": "assistant"}]
                    if new_history:
                        logger.info("Resolving...")
                        resolve_response(new_history)
                        schedule_refresh_assistant()
                        logger.info("Resolved")
                    if interrupt_event.is_set():
                        logger.info("Interrupted")
                        interrupt_event.clear()

                    transcript = []
                    timestamps = []
                    while not audio_queue.empty():
                        audio_queue.get()
                    while not text_queue.empty():
                        text_queue.get()
                    speaking.clear()
                    play_audio_file(core_path+"/tone_one.wav", blocking=True)
    logger.info("Converse trying to shutdown")
    interrupt_event.set()
    speaking.set()
    multiprocessing_stop_event.set()
    audio_queue.put((None, time.time()))
    audio_queue.put((None, time.time()))
    audio_queue.put((None, time.time()))
    logger.info("Capture trying to shutdown")
    capture_process.join(timeout=10)
    if capture_process.is_alive():
        logger.warning("Terminating Capture...")
    capture_process.terminate()
    logger.info("Transcribe trying to shutdown")
    text_process.join(timeout=10)
    if text_process.is_alive():
        logger.warning("Terminating Transcribe...")
    capture_process.terminate()


if __name__ == "__main__":
    converse(60, threading.Event(), threading.Event(), threading.Event())



import subprocess
import sys
import os
import uuid
import io
import re

if sys.platform != "darwin" and sys.platform != "linux":
    from pydub import AudioSegment
    from gtts import gTTS

if not os.path.exists(os.path.join(os.path.expanduser("~"), "Jarvis Logs")):
    os.mkdir(os.path.join(os.path.expanduser("~"), "Jarvis Logs"))
audio_folder = os.path.join(os.path.expanduser("~"), "Jarvis Logs", "temp_audio")
if not os.path.exists(audio_folder):
    os.mkdir(audio_folder)

if sys.platform == 'darwin':
    out = subprocess.run(['say', '-v', '?'], capture_output=True)
    vflag = []
    if out.stdout.decode("utf-8").find("Samantha (Enhanced)") >= 0:
        vflag = ['-v', 'Samantha (Enhanced)']
    if out.stdout.decode("utf-8").find("Tom (Enhanced)") >= 0:
        vflag = ['-v', 'Tom (Enhanced)']
    if out.stdout.decode("utf-8").find("Evan (Enhanced)") >= 0:
        vflag = ['-v', 'Evan (Enhanced)']
if sys.platform == 'linux':
    env = os.environ.copy()
    env["PATH"] = sys.executable + os.pathsep + env["PATH"]


class TextToSpeechError(Exception):
    """
    Exception raised when the text to speech conversion fails.
    """
    def __init__(self, sentence):
        """
        Initialise the exception.

        :param sentence: str, The sentence that failed to be converted to speech.
        :type sentence: str
        """
        self.sentence = sentence
        super().__init__(f"The following sentence was too long to turn into voice '{self.sentence}'.")


def simplify_urls(text):
    """
    Simplify URLs in the given text by removing the protocol and "www." and anything after the domain name.
    :param text: The text to simplify URLs in.
    :type text: str
    :return: The modified text.
    :rtype: str
    """
    # Regular expression to match URLs
    url_pattern = r'http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\\(\\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+'

    # Find all URLs in the text
    urls = re.findall(url_pattern, text)

    # Function to extract and simplify the domain name from a URL
    def get_simplified_domain(url):
        domain = re.sub(r'(http[s]?://|www\.)', '', url)  # Remove the protocol and "www."
        domain = re.sub(r'(/.*)', '', domain)            # Remove anything after the domain
        return domain

    # Replace each URL with its simplified domain name
    for url in urls:
        simplified_domain = get_simplified_domain(url)
        text = text.replace(url, simplified_domain)

    return text


def find_longest_sentence(text):
    """
    Find the longest sentence in the given text.
    :param text: The text to find the longest sentence in.
    :type text: str
    :return: The longest sentence.
    :rtype: str
    """
    # Split the text into sentences using regex
    sentences = re.split(r' *[\.\?!][\'"\)\]]* *', text)

    # Find the longest sentence
    longest_sentence = max(sentences, key=len)

    return longest_sentence


def split_longest_sentence(text):
    """
    Split the longest sentence in the given text into smaller sentences.
    :param text: The text to split the longest sentence in.
    :type text: str
    :return: The modified text.
    :rtype: str
    """
    # Find the longest sentence
    longest_sentence = find_longest_sentence(text)

    # Split the longest sentence into chunks
    chunks = split_sentence(longest_sentence)

    # Replace the longest sentence in the text with the smaller sentences
    new_longest_sentence = '. '.join(chunks)
    modified_text = text.replace(longest_sentence, new_longest_sentence)

    return modified_text


def capitalize_first_letter(sentence: str) -> str:
    """
    Capitalize the first letter of the given sentence.

    :param sentence: The sentence to capitalize the first letter.
    :type sentence: str
    :return: The sentence with the first letter capitalized.
    :rtype: str
    """
    if len(sentence) > 0:
        return sentence[0].upper() + sentence[1:]
    return sentence


def remove_trailing_comma(sentence: str) -> str:
    """
    Remove trailing comma from the given sentence.

    :param sentence: The sentence to remove the trailing comma.
    :type sentence: str
    :return: The sentence without the trailing comma.
    :rtype: str
    """
    return sentence.rstrip(',')


def find_best_split(sentence: str) -> int:
    """
    Find the best split index for a given sentence, considering commas, colons, and semicolons.

    :param sentence: The input sentence to find the best split index.
    :return: The best split index or None if there are no delimiters.
    :rtype: int or None
    """
    # Find the indices of all delimiters
    delimiter_indices = [m.start() for m in re.finditer(r'[:,;]', sentence)]

    if not delimiter_indices:
        return None

    # Calculate the lengths of the two halves for each delimiter
    half_lengths = [(abs(len(sentence[:i]) - len(sentence[i+1:])), i) for i in delimiter_indices]

    # Find the delimiter that results in the most evenly-sized halves
    best_split_index = min(half_lengths, key=lambda x: x[0])[1]

    return best_split_index


def split_sentence(sentence: str) -> list:
    """
    Split the input sentence into two parts at the best delimiter.

    :param sentence: The input sentence to split.
    :type sentence: str
    :return: A list of two sentences after splitting.
    :rtype: list
    """
    best_split_index = find_best_split(sentence)

    if best_split_index is None:
        return [sentence]

    first_half = sentence[:best_split_index].strip()
    second_half = sentence[best_split_index + 1:].strip()

    # Capitalize the first letter and remove trailing commas
    first_half = capitalize_first_letter(remove_trailing_comma(first_half))
    second_half = capitalize_first_letter(remove_trailing_comma(second_half))

    return [first_half, second_half]


def text_to_speech(text: str, model="gpt-4", stream=False):
    """
    Convert the given text to speech using the specified model.

    :param text: The text to convert to speech.
    :type text: str
    :param model: The model to use for text-to-speech.
    :type model: str
    :param stream: Whether to return the audio content as a stream.
    :type stream: bool
    :return: The path to the audio file or the audio content as a stream.
    :rtype: str or bytes
    """
    slow_flag = True
    if model.find("gpt-4") >= 0:
        slow_flag = False
    if sys.platform == 'darwin':
        fixed_text = text.replace('"', r'\"')
        pitch = "44"
        if slow_flag:
            pitch = "40"
        first_word = fixed_text.split(" ")[0]
        rest_of_text = fixed_text.replace(first_word, "")
        fixed_text = "[[rate 175]] " + first_word + "[[rate 200]] " + rest_of_text
        text_cmd = f'[[pbas {pitch}]] [[slnc 300]]{fixed_text}[[slnc 200]]'
        output_file = os.path.join(audio_folder, str(uuid.uuid4()) + ".wav")
        result = subprocess.run(['say']+vflag+[text_cmd, "-o", output_file, '--data-format=LEI16@22050'],
                                capture_output=True)
        if not stream:
            return output_file
        if result.returncode != 0:
            raise Exception("Say command error: " + result.stderr.decode("utf-8"))
        with open(output_file, 'rb') as file:
            file.seek(0)
            byte_data = file.read()
        os.remove(output_file)
        return byte_data
    if sys.platform == 'linux':
        print(text)
        fixed_text = text.replace('"', r'\"')
        speed = ".85"
        if slow_flag:
            speed = "1"
        text_cmd = fixed_text
        output_file = os.path.join(audio_folder, str(uuid.uuid4()) + ".wav")
        print(['mimic3', text_cmd, ">", output_file])
        result = subprocess.run(f"mimic3 \"{text_cmd}\" --length-scale {speed} > '{output_file}'",
                                capture_output=True,
                                env=env, shell=True)
        if not stream:
            return output_file
        if result.returncode != 0:
            raise Exception("Say command error: " + result.stderr.decode("utf-8"))
        with open(output_file, 'rb') as file:
            file.seek(0)
            byte_data = file.read()
        os.remove(output_file)
        return byte_data
    else:
        tts = gTTS(text, lang='en', slow=slow_flag)

        if stream:
            mp3_buffer = io.BytesIO()
            tts.write_to_fp(mp3_buffer)
            mp3_buffer.seek(0)
            mp3_segment = AudioSegment.from_file(mp3_buffer, format="mp3")
            wav_buffer = io.BytesIO()
            mp3_segment.export(wav_buffer, format="wav")
            wav_buffer.seek(0)
            audio_content = wav_buffer.read()
            return audio_content
        else:
            output_file = os.path.join(audio_folder, str(uuid.uuid4()) + ".wav")
            mp3_output = io.BytesIO()
            tts.write_to_fp(mp3_output)
            mp3_output.seek(0)
            mp3_segment = AudioSegment.from_file(mp3_output, format="mp3")
            mp3_segment.export(output_file, format="wav")
            return output_file



import multiprocessing
import os
import whisper
from torch.cuda import is_available
from io import BytesIO
from soundfile import read
from numpy import float32
import warnings

from .speaker_recognition import recognition_process
from .logger_config import get_log_folder_path, get_logger
logger = get_logger()


warnings.filterwarnings("ignore", message="FP16 is not supported on CPU; using FP32 instead")

dir_root = get_log_folder_path()

dir_path = os.path.join(dir_root, "whisper_models")
if not os.path.exists(dir_path):
    os.mkdir(dir_path)

if is_available():
    model = None
    logger.info("Using GPU! (if using just module need to initialize module where you want it)")
else:
    model = whisper.load_model("base.en", download_root=dir_path)


def convert_to_text(audio: BytesIO, text_only=False) -> str or list:
    """
    Convert the given audio data to text using speech recognition.

    :param audio: The audio data to convert.
    :type audio: BytesIO
    :param text_only: Whether to return only the text or the full result.
    :type text_only: bool
    :return: str, the recognized text from the audio
    :rtype: str or list
    """
    # Read audio data
    audio.seek(0)
    array_audio, sampling_rate = read(audio)
    array_audio = array_audio.astype(float32)
    result = model.transcribe(array_audio)

    # Return the result
    if text_only:
        return result["text"]
    else:
        return result


def fuse_text_and_speakers(whisper_output, speakers):
    segments = whisper_output['segments']
    annotated_text = ""
    last_speaker = None
    segment_index = 0

    # Function to append text with the speaker label
    def append_with_speaker(text, speaker):
        nonlocal annotated_text, last_speaker
        if speaker != last_speaker:
            annotated_text += f"\n[{speaker}]: " if text.strip() else ""
            last_speaker = speaker
        annotated_text += text + " "

    # Iterate through each segment and assign speakers
    while segment_index < len(segments):
        segment = segments[segment_index]
        overlapping_speakers = [speaker_label for start, end, speaker_label in speakers if
                                start <= segment['end'] and end >= segment['start']]

        if not overlapping_speakers:
            append_with_speaker(segment['text'], 'Unknown Speaker')
        elif len(overlapping_speakers) == 1:
            append_with_speaker(segment['text'], overlapping_speakers[0])
        else:
            append_with_speaker(segment['text'], 'Multiple Speakers')

        segment_index += 1

    return annotated_text.strip()


def audio_processing_thread(audio_queue, text_queue, speaking, stop_event):
    """
    Transcribes audio from the audio queue and puts it in the text queue.
    :param audio_queue:
    :param text_queue:
    :param speaking:
    :param stop_event:
    :return:
    """
    try:
        recog_audio_queue = multiprocessing.Queue()
        speakers_queue = multiprocessing.Queue()
        recog_process = multiprocessing.Process(target=recognition_process,
                                               args=(recog_audio_queue, speakers_queue, speaking,
                                                     stop_event), )

        recog_process.start()
        if is_available():
            global model
            model = whisper.load_model("base.en", download_root=dir_path)
        while stop_event.is_set() is False:
            if audio_queue.empty():
                speaking.wait(timeout=0.2)
            else:
                audio_data, ts = audio_queue.get()
                if audio_data is None:
                    text_queue.put(("", ts))
                elif not speaking.is_set() and not stop_event.is_set():
                    while not recog_audio_queue.empty():
                        recog_audio_queue.get()
                    while not speakers_queue.empty():
                        speakers_queue.get()
                    if not speaking.is_set() and not stop_event.is_set():
                        recog_audio_queue.put(audio_data)
                        text = convert_to_text(audio_data)
                        if not speaking.is_set() and not stop_event.is_set():
                            try:
                                speakers = speakers_queue.get(timeout=5)
                                text = fuse_text_and_speakers(text, speakers)
                            except multiprocessing.queues.Empty:
                                text = text['text']

                            if not speaking.is_set() and not stop_event.is_set():
                                text_queue.put((text, ts))
    except KeyboardInterrupt:
        return


# # Example usage
# whisper_output = {
#     "segments": [
#         {"start": 0.0, "end": 3.0, "text": "Hello, how are you?"},
#         {"start": 3.0, "end": 6.0, "text": "I'm fine, thank you."}
#     ]
# }
# whisper_output = {
#     "segments": [
#         {"start": 0.0, "end": 2.0, "text": "Good morning, everyone."},
#         {"start": 2.0, "end": 4.0, "text": "Today, we are discussing our new project."},
#         {"start": 4.0, "end": 6.0, "text": "I think it's a great opportunity to grow."},
#         {"start": 6.0, "end": 8.0, "text": "Absolutely, and with the right strategy, we can succeed."},
#         {"start": 8.0, "end": 10.0, "text": "Let's not forget about the challenges ahead."},
#         {"start": 10.0, "end": 12.0, "text": "Yes, especially the tight deadlines."},
#         {"start": 12.0, "end": 12.5, "text": "We'll need to collaborate effectively."},
#         {"start": 14.0, "end": 16.0, "text": "I agree, teamwork is key here."}
#     ]
# }
#
# speakers = [
#     (0.0, 3.0, 'SPEAKER_01'),  # Overlaps with SPEAKER_02
#     (2.0, 5.0, 'SPEAKER_02'),  # Overlaps with SPEAKER_01 and SPEAKER_03
#     (4.0, 7.0, 'SPEAKER_03'),  # Overlaps with SPEAKER_02
#     (7.0, 9.0, 'SPEAKER_04'),
#     (9.0, 11.0, 'SPEAKER_05'),
#     # Omitting explicit 'Unknown Speaker' label
#     (13.0, 15.0, 'SPEAKER_06')
#     # No speaker data for segments from 11.0 to 13.0 and the last segment
# ]
#
# annotated_text = fuse_text_and_speakers(whisper_output, speakers)
# print(annotated_text)



import sys
if sys.platform == 'linux':
    __import__('pysqlite3')
    sys.modules['sqlite3'] = sys.modules.pop('pysqlite3')
__version__ = '0.3.2'
# Path: src/jarvis_conversationalist/conversationalist.py



import os
import time
import threading
import soundfile as sf
from numpy import linspace, int16, sqrt, maximum, mean, square, frombuffer


def play_audio_file(file_path, blocking: bool = True, loops=1, delay: float = 0, destroy=False,
                    added_stop_event: threading.Event = None) -> threading.Event:
    """
    Play an audio file using pyaudio.

    :param file_path: path to the audio file or list of paths to play in sequence
    :type file_path: str or list of str
    :param blocking: whether the audio playback should block the main thread (default: True)
    :type blocking: bool
    :param loops: the number of times to loop the audio file (default: 1) or list of loop counts for each file
    :type loops: int or list of int
    :param delay: the delay in seconds before starting playback (default: 0)
    :type delay: float
    :param destroy: whether to destroy the file after playback (default: False) or list of destroy flags for each file
    :type destroy: bool or list of bool
    :param added_stop_event: an event to signal stopping the playback (only for non-blocking mode)
    :type added_stop_event: threading.Event
    :return: an event to signal stopping the playback (only for non-blocking mode)
    :rtype: threading.Event
    """
    stop_event = threading.Event()

    if blocking:
        time.sleep(delay)
        _play_audio_file_blocking(file_path, stop_event, loops, 0, destroy, added_stop_event)
    else:
        playback_thread = threading.Thread(target=_play_audio_file_blocking,
                                           args=(file_path, stop_event, loops, delay, destroy, added_stop_event))
        playback_thread.start()

    return stop_event


def _play_audio_file_blocking(file_path: str, stop_event: threading.Event, loops: int, delay: float, destroy: bool,
                              added_stop_event: threading.Event):
    """
    Play an audio file using pyaudio, blocking the calling thread until playback is complete or stopped.

    :param file_path: path to the audio file
    :type file_path: str
    :param stop_event: an event to signal stopping the playback
    :type stop_event: threading.Event
    :param loops: the number of times to loop the audio file
    :type loops: int
    :param delay: the delay in seconds before starting playback
    :type delay: float
    :param destroy: whether to destroy the file after playback
    :type destroy: bool
    :param added_stop_event: an event to signal stopping the playback
    :type added_stop_event: threading.Event
    """
    if isinstance(file_path, list):
        files_updated = []
        for file in file_path:
            files_updated.append(file)
        file_path = files_updated
        if isinstance(loops, list) and isinstance(destroy, list):
            for i, file in enumerate(file_path):
                _play_audio_file_blocking(file, stop_event, loops[i], delay, destroy[i], added_stop_event)
        elif isinstance(loops, list):
            for i, file in enumerate(file_path):
                _play_audio_file_blocking(file, stop_event, loops[i], delay, destroy, added_stop_event)
        else:
            for file in file_path:
                _play_audio_file_blocking(file, stop_event, loops, delay, destroy, added_stop_event)
        return
    else:
        chunk = 8192
        if isinstance(loops, list):
            loops = loops[0]  # Assumption: If loops is a list, take the first value

        # Wait for the specified delay
        if added_stop_event:
            added_stop_event.wait(timeout=delay)
        else:
            stop_event.wait(timeout=delay)
        import sounddevice as sd
        # Play the audio file
        for loop in range(loops):
            if not stop_event.is_set() or (added_stop_event and not added_stop_event.is_set()):
                data, fs = sf.read(file_path)
                sd.play(data, fs)
                while sd.get_stream().active:
                    if stop_event.is_set() or (added_stop_event and added_stop_event.is_set()):
                        sd.stop()
                        break
                    if added_stop_event:
                        added_stop_event.wait(timeout=.02)
                    else:
                        stop_event.wait(timeout=.02)
                sd.stop()
        # Destroy the file if needed
        if destroy:
            os.remove(file_path)


def fade_out(data: bytes, fade_duration: int, rms_threshold: int = 1000) -> bytes:
    """
    Fade out the audio data.

    :param data: the audio data to fade out
    :type data: bytes
    :param fade_duration: the duration of the fade in samples
    :type fade_duration: int
    :param rms_threshold: the threshold for determining if audio is present
    :type rms_threshold: int
    :return: the audio data with a fade-out applied
    :rtype: bytes
    """

    def rms(aud_data):
        return sqrt(maximum(mean(square(aud_data)), 0))

    num_samples = len(data) // 2  # Divide by 2 for 16-bit audio samples
    fade_samples = min(num_samples, fade_duration)
    audio_data = frombuffer(data, dtype=int16).copy()  # Create a writeable copy of the array

    start_fade = 0
    for i in range(0, num_samples - fade_samples, fade_samples):
        if rms(audio_data[i:i + fade_samples]) < rms_threshold:
            start_fade = i
            break

    fade = linspace(1, 0, num_samples - start_fade).astype(int16)  # Convert fade array to int16
    audio_data[start_fade:] *= fade
    return audio_data.tobytes()



import os
import uuid
import warnings
from io import BytesIO

warnings.filterwarnings("ignore", message=".*audio._backend.set_audio_backend.*")

from torch import device
from torch.cuda import is_available
from pyannote.audio import Pipeline
from .audio_vectordb import LocalAudioDB


def db_speaker_id(speaker_id):
    """
    This method converts a speaker ID to the format used in the speaker databases.
    """
    speaker_id = "Unknown Speaker " + str(int(speaker_id.replace("Unknown Speaker ", ""))).zfill(9)
    return speaker_id


def pretty_speaker_id(speaker_id):
    """
    This method converts a speaker ID to a more readable format.
    """
    speaker_id = "Unknown Speaker " + str(int(speaker_id.replace("Unknown Speaker ", "")))
    return speaker_id


def pretty_known_speaker_id(speaker_id):
    """
    This method converts a speaker ID to a more readable format.
    """
    speaker_id = speaker_id.split(" -- ")[0]
    return speaker_id


class SpeakerIdentifier:
    """
    This class is responsible for identifying speakers in an audio stream. It uses a pretrained model from Pyannote
     for speaker diarization and stores known and unknown speaker embeddings in separate databases.
    """

    def __init__(
        self,
        persist_directory: str = "speaker_database",
        similar_speaker_threshold: float = .5,
    ):
        """
        The constructor for the SpeakerIdentifier class. It initializes the class with a directory to persist speaker
        data and a threshold for determining if a speaker is similar to a known speaker.
        """
        self.similar_speaker_threshold = similar_speaker_threshold
        self.known_speakers_path = os.path.join(persist_directory, "known_speakers_db")
        self.known_speakers = LocalAudioDB(self.known_speakers_path)
        self.unknown_speakers_path = os.path.join(persist_directory, "unknown_speakers_db")
        self.unknown_speakers = LocalAudioDB(self.unknown_speakers_path)
        self.pipeline = Pipeline.from_pretrained("pyannote/speaker-diarization-3.1",
                                                 use_auth_token="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"
                                                 )

    def speedup_if_able(self):
        if is_available():
            self.pipeline.to(device("cuda"))

    def get_next_unknown_speaker_id(self):
        """
        This method retrieves the next available ID for an unknown speaker from the unknown speakers database.

        :return: the speaker ID
        :rtype: str
        """
        largest_id = self.unknown_speakers.get_largest_id()
        if largest_id is None:
            largest_id = "Unknown Speaker 0"
        largest_id = "Unknown Speaker "+str(int(largest_id.replace("Unknown Speaker ", ""))+1)
        return largest_id

    def get_last_unknown_speaker_id(self):
        """
        This method retrieves the last added ID for an unknown speaker from the unknown speakers database.

        :return: the speaker ID
        :rtype: str
        """
        largest_id = self.unknown_speakers.get_largest_id()
        if largest_id is None:
            largest_id = "Unknown Speaker 0"
        largest_id = "Unknown Speaker "+str(int(largest_id.replace("Unknown Speaker ", "")))
        return largest_id

    def check_known_speakers(self, embedding):
        """
        This method checks if a given speaker embedding is similar to any known speaker embeddings.
        If a similar speaker is found, their ID is returned.

        :param embedding: the speaker embedding to check
        :type embedding: numpy array
        :return: the speaker ID
        :rtype: str
        """
        closest_id, closest_embedding, closest_distance = self.known_speakers.find_closest_embedding(embedding)
        if closest_distance:
            if float(closest_distance) < self.similar_speaker_threshold:
                closest_id = pretty_known_speaker_id(closest_id)
                return closest_id
        return None

    def add_known_speaker(self, embedding, name, new=False):
        """
        This method adds a known speaker to the known speakers database.

        :param embedding: the speaker embedding
        :type embedding: numpy array
        :param name: the speaker name
        :type name: str
        :param new: whether this is a new speaker
        :type new: bool
        """
        name = name.strip()+" -- "+str(uuid.uuid4())
        if new:
            result = self.known_speakers.query_embeddings(name)
            if result:
                raise Exception("Speaker name already exists.")
        self.known_speakers.add_embedding(name, embedding)

    def remove_known_speaker(self, name):
        """
        This method removes a known speaker from the known speakers' database.

        :param name: the speaker name
        :type name: str
        """
        results = self.known_speakers.query_embeddings(name + " -- ")
        if results:
            for result in results:
                self.known_speakers.remove_embedding(result[0])
        else:
            raise Exception("Speaker name does not exist.")

    def get_add_unknown_speaker(self, embedding):
        """
        This method checks if a given speaker embedding is similar to any unknown speaker embeddings.
        If a similar speaker is found, their ID is returned. If not, a new speaker ID is generated and the embedding
        is added to the unknown speakers' database.

        :param embedding: the speaker embedding to check
        :type embedding: numpy array
        :return: the speaker ID
        :rtype: str
        """
        closest_id, closest_embedding, closest_distance = self.unknown_speakers.find_closest_embedding(embedding)
        if closest_distance is not None:
            if float(closest_distance) < self.similar_speaker_threshold:
                return pretty_speaker_id(closest_id)
        speaker_id = self.get_next_unknown_speaker_id()
        self.unknown_speakers.add_embedding(db_speaker_id(speaker_id), embedding)
        return pretty_speaker_id(speaker_id)

    def get_unknown_embedding(self, speaker_id):
        """
        This method retrieves the embedding of an unknown speaker from the unknown speakers' database.

        :param speaker_id: the speaker ID
        :type speaker_id: str
        :return: the speaker embedding
        :rtype: numpy array
        """
        id, embedding = self.unknown_speakers.get_embedding(db_speaker_id(speaker_id))
        return embedding

    def remove_unknown_speaker(self, speaker_id):
        """
        This method removes an unknown speaker from the unknown speakers database.

        :param speaker_id: the speaker ID
        :type speaker_id: str
        """
        self.unknown_speakers.remove_embedding(db_speaker_id(speaker_id))

    def get_speakers(self, audio_data_io):
        """
        This method takes an audio data stream as input and returns a timeline of speakers.
        It uses the Pyannote pipeline to perform speaker diarization and retrieve speaker embeddings.
        These embeddings are then checked against the known and unknown speaker databases to identify the speakers.

        :param audio_data_io: the audio data stream to process
        :type audio_data_io: BytesIO
        :return: the timeline of speakers
        :rtype: list
        """
        audio_data_io.seek(0)

        # Process the NumPy array with Pyannote
        diarization, embeddings = self.pipeline(audio_data_io, return_embeddings=True)

        # combine embeddings with diarization.labels() into dict
        speakers = {}
        for i, speaker in enumerate(diarization.labels()):
            speakers[speaker] = embeddings[i]

        # Create a timeline
        timeline = []
        found_speakers = {}
        for turn, _, speaker_code in diarization.itertracks(yield_label=True):
            start, end = turn.start, turn.end
            if speaker_code in found_speakers:
                continue
            else:
                speaker_id = self.check_known_speakers(speakers[speaker_code])
                if speaker_id is None:
                    speaker_id = self.get_add_unknown_speaker(speakers[speaker_code])
                found_speakers[speaker_code] = speaker_id
            timeline.append((start, end, speaker_id))
        return timeline

